"""Interface layer for pykube library shared between Galaxy and Pulsar."""

import logging
import os
import re
from pathlib import PurePath

try:
    from pykube.config import KubeConfig
    from pykube.exceptions import HTTPError
    from pykube.http import HTTPClient
    from pykube.objects import (
        Ingress,
        Job,
        Pod,
        Service,
    )
except ImportError as exc:
    KubeConfig = None
    Ingress = None
    Job = None
    Pod = None
    Service = None
    HTTPError = None
    K8S_IMPORT_MESSAGE = (
        "The Python pykube package is required to use "
        "this feature, please install it or correct the "
        f"following error:\nImportError {exc}"
    )

log = logging.getLogger(__name__)

DEFAULT_JOB_API_VERSION = "batch/v1"
DEFAULT_SERVICE_API_VERSION = "v1"
DEFAULT_INGRESS_API_VERSION = "networking.k8s.io/v1"
DEFAULT_NAMESPACE = "default"
INSTANCE_ID_INVALID_MESSAGE = (
    "Galaxy instance [%s] is either too long "
    "(>20 characters) or it includes non DNS "
    "acceptable characters, ignoring it."
)


def ensure_pykube():
    if KubeConfig is None:
        raise Exception(K8S_IMPORT_MESSAGE)


def pykube_client_from_dict(params):
    if "k8s_use_service_account" in params and params["k8s_use_service_account"]:
        pykube_client = HTTPClient(KubeConfig.from_service_account())
    else:
        config_path = params.get("k8s_config_path")
        if config_path is None:
            config_path = os.environ.get("KUBECONFIG", None)
        if config_path is None:
            config_path = "~/.kube/config"
        pykube_client = HTTPClient(KubeConfig.from_file(config_path))
    return pykube_client


def produce_k8s_job_prefix(app_prefix=None, instance_id=None):
    job_name_elems = [app_prefix or "", instance_id or ""]
    return "-".join(elem for elem in job_name_elems if elem)


def pull_policy(params):
    # If this doesn't validate it returns None, that seems odd?
    if "k8s_pull_policy" in params:
        if params["k8s_pull_policy"] in ["Always", "IfNotPresent", "Never"]:
            return params["k8s_pull_policy"]
    return None


def find_service_object_by_name(pykube_api, service_name, namespace=None):
    if not service_name:
        raise ValueError("service name must not be empty")
    return Service.objects(pykube_api).filter(field_selector={"metadata.name": service_name}, namespace=namespace)


def find_ingress_object_by_name(pykube_api, ingress_name, namespace=None):
    if not ingress_name:
        raise ValueError("ingress name must not be empty")
    return Ingress.objects(pykube_api).filter(field_selector={"metadata.name": ingress_name}, namespace=namespace)


def find_job_object_by_name(pykube_api, job_name, namespace=None):
    if not job_name:
        raise ValueError("job name must not be empty")
    return Job.objects(pykube_api).filter(field_selector={"metadata.name": job_name}, namespace=namespace)


def find_pod_object_by_name(pykube_api, job_name, namespace=None):
    return Pod.objects(pykube_api).filter(selector=f"job-name={job_name}", namespace=namespace)


def is_pod_unschedulable(pykube_api, pod, namespace=None):
    is_unschedulable = any(c.get("reason") == "Unschedulable" for c in pod.obj["status"].get("conditions", []))
    if pod.obj["status"].get("phase") == "Pending" and is_unschedulable:
        return True

    return False


def delete_job(job, cleanup="always"):
    job_failed = job.obj["status"]["failed"] > 0 if "failed" in job.obj["status"] else False
    # Scale down the job just in case even if cleanup is never
    job.scale(replicas=0)
    api_delete = cleanup == "always"
    if not api_delete and cleanup == "onsuccess" and not job_failed:
        api_delete = True
    if api_delete:
        delete_options = {"apiVersion": "v1", "kind": "DeleteOptions", "propagationPolicy": "Background"}
        r = job.api.delete(json=delete_options, **job.api_kwargs())
        job.api.raise_for_status(r)


def delete_ingress(ingress, cleanup="always", job_failed=False):
    api_delete = cleanup == "always"
    if not api_delete and cleanup == "onsuccess" and not job_failed:
        api_delete = True
    if api_delete:
        delete_options = {"apiVersion": "v1", "kind": "DeleteOptions", "propagationPolicy": "Background"}
        r = ingress.api.delete(json=delete_options, **ingress.api_kwargs())
        ingress.api.raise_for_status(r)


def delete_service(service, cleanup="always", job_failed=False):
    api_delete = cleanup == "always"
    if not api_delete and cleanup == "onsuccess" and not job_failed:
        api_delete = True
    if api_delete:
        delete_options = {"apiVersion": "v1", "kind": "DeleteOptions", "propagationPolicy": "Background"}
        r = service.api.delete(json=delete_options, **service.api_kwargs())
        service.api.raise_for_status(r)


def job_object_dict(params, job_prefix, spec):
    k8s_job_obj = {
        "apiVersion": params.get("k8s_job_api_version", DEFAULT_JOB_API_VERSION),
        "kind": "Job",
        "metadata": {
            "generateName": f"{job_prefix}-",
            "namespace": params.get("k8s_namespace", DEFAULT_NAMESPACE),
        },
        "spec": spec,
    }
    return k8s_job_obj


def service_object_dict(params, service_name, spec):
    k8s_service_obj = {
        "apiVersion": params.get("k8s_service_api_version", DEFAULT_SERVICE_API_VERSION),
        "kind": "Service",
        "metadata": {
            "name": service_name,
            "namespace": params.get("k8s_namespace", DEFAULT_NAMESPACE),
        },
    }
    k8s_service_obj["metadata"].update(spec.pop("metadata", {}))
    k8s_service_obj.update(spec)
    return k8s_service_obj


def ingress_object_dict(params, ingress_name, spec):
    k8s_ingress_obj = {
        "apiVersion": params.get("k8s_ingress_api_version", DEFAULT_INGRESS_API_VERSION),
        "kind": "Ingress",
        "metadata": {
            "name": ingress_name,
            "namespace": params.get("k8s_namespace", DEFAULT_NAMESPACE),
            # TODO: Add default annotations
        },
    }
    k8s_ingress_obj["metadata"].update(spec.pop("metadata", {}))
    k8s_ingress_obj.update(spec)
    return k8s_ingress_obj


def parse_pvc_param_line(pvc_param):
    """
    :type pvc_param: str
    :param pvc_param: the pvc mount param in the format ``pvc-name/subpath/desired:/mountpath/desired[:r]``

    :rtype: dict
    :return: a dict
      like::

        {"name": "pvc-name",
         "subPath": "subpath/desired",
         "mountPath": "/mountpath/desired",
         "readOnly": False}
    """
    claim, _, rest = pvc_param.partition(":")
    mount_path, _, mode = rest.partition(":")
    read_only = mode == "r"
    claim_name, _, subpath = claim.partition("/")
    return {
        "name": claim_name.strip(),
        "subPath": subpath.strip(),
        "mountPath": mount_path.strip(),
        "readOnly": read_only,
    }


def generate_relative_mounts(pvc_param, files):
    """
    Maps a list of files as mounts, relative to the base volume mount.
    For example, given the pvc mount:
    {
        'name': 'my_pvc',
        'mountPath': '/galaxy/database/jobs',
        'subPath': 'data',
        'readOnly': False
    }

    and files: ['/galaxy/database/jobs/01/input.txt', '/galaxy/database/jobs/01/working']

    returns each file as a relative mount as follows:
    [
        {
          'name': 'my_pvc',
          'mountPath': '/galaxy/database/jobs/01/input.txt',
          'subPath': 'data/01/input.txt',
          'readOnly': False
        },
        {
          'name': 'my_pvc',
          'mountPath': '/galaxy/database/jobs/01/working',
          'subPath': 'data/01/working',
          'readOnly': False
        }
    ]

    :param pvc_param: the pvc claim dict
    :param files: a list of file or folder names
    :return: A list of volume mounts
    """
    if not pvc_param:
        return
    param_claim = parse_pvc_param_line(pvc_param)
    claim_name = param_claim["name"]
    base_subpath = PurePath(param_claim.get("subPath", ""))
    base_mount = PurePath(param_claim["mountPath"])
    read_only = param_claim["readOnly"]
    volume_mounts = []
    for f in files:
        file_path = PurePath(str(f))
        if base_mount not in file_path.parents:
            # force relative directory, needed for the job working directory in particular
            file_path = base_mount.joinpath(file_path.relative_to("/") if file_path.is_absolute() else file_path)
        relpath = file_path.relative_to(base_mount)
        subpath = base_subpath.joinpath(relpath)
        volume_mounts.append(
            {"name": claim_name, "mountPath": str(file_path), "subPath": str(subpath), "readOnly": read_only}
        )
    return volume_mounts


def deduplicate_entries(obj_list):
    # remove duplicate entries in a list of dictionaries
    # based on: https://stackoverflow.com/a/9428041
    return [i for n, i in enumerate(obj_list) if i not in obj_list[n + 1 :]]


def get_volume_mounts_for_job(job_wrapper, data_claim=None, working_claim=None):
    volume_mounts = []
    if data_claim:
        volume_mounts.extend(generate_relative_mounts(data_claim, job_wrapper.job_io.get_input_fnames()))
        # for individual output files, mount the parent folder of each output as there could be wildcard outputs
        output_folders = deduplicate_entries(
            [str(PurePath(str(f)).parent) for f in job_wrapper.job_io.get_output_fnames()]
        )
        volume_mounts.extend(generate_relative_mounts(data_claim, output_folders))

    if working_claim:
        volume_mounts.extend(generate_relative_mounts(working_claim, [job_wrapper.working_directory]))

    return deduplicate_entries(volume_mounts)


def galaxy_instance_id(params):
    """Parse and validate the id of the Galaxy instance from supplied dict.

    This will be added to Jobs and Pods names, so it needs to be DNS friendly,
    this means: `The Internet standards (Requests for Comments) for protocols mandate that component hostname labels
    may contain only the ASCII letters 'a' through 'z' (in a case-insensitive manner), the digits '0' through '9',
    and the minus sign ('-').`

    It looks for the value set on params['k8s_galaxy_instance_id'], which might or not be set. The
    idea behind this is to allow the Galaxy instance to trust (or not) existing k8s Jobs and Pods that match the
    setup of a Job that is being recovered or restarted after a downtime/reboot.
    """
    if "k8s_galaxy_instance_id" in params:
        raw_value = params["k8s_galaxy_instance_id"]
        if re.match(r"(?!-)[a-z\d-]{1,20}(?<!-)$", raw_value):
            return raw_value
        else:
            log.error(INSTANCE_ID_INVALID_MESSAGE % raw_value)
    return None


__all__ = (
    "DEFAULT_JOB_API_VERSION",
    "DEFAULT_SERVICE_API_VERSION",
    "DEFAULT_INGRESS_API_VERSION",
    "ensure_pykube",
    "find_service_object_by_name",
    "find_ingress_object_by_name",
    "find_job_object_by_name",
    "find_pod_object_by_name",
    "galaxy_instance_id",
    "HTTPError",
    "is_pod_unschedulable",
    "Job",
    "Service",
    "Ingress",
    "job_object_dict",
    "service_object_dict",
    "ingress_object_dict",
    "Pod",
    "produce_k8s_job_prefix",
    "pull_policy",
    "pykube_client_from_dict",
    "delete_job",
    "delete_service",
    "delete_ingress",
    "get_volume_mounts_for_job",
    "parse_pvc_param_line",
)



#!/usr/bin/env python
"""
Runs SRMA on a SAM/BAM file;
TODO: more documentation

usage: srma_wrapper.py [options]

See below for options
"""

import optparse
import os
import shutil
import subprocess
import sys
import tempfile


def stop_err(msg):
    sys.stderr.write("%s\n" % msg)
    sys.exit()


def parseRefLoc(refLoc, refUID):
    for line in open(refLoc):
        if not line.startswith("#"):
            fields = line.strip().split("\t")
            if len(fields) >= 3:
                if fields[0] == refUID:
                    return fields[1]
    return None


def __main__():
    parser = optparse.OptionParser()
    parser.add_option("-r", "--ref", dest="ref", help="The reference genome to index and use")
    parser.add_option("-u", "--refUID", dest="refUID", help="The pre-index reference genome unique Identifier")
    parser.add_option("-i", "--input", dest="input", help="The SAM/BAM input file")
    parser.add_option("-I", "--inputIndex", dest="inputIndex", help="The SAM/BAM input index file")
    parser.add_option("-o", "--output", dest="output", help="The SAM/BAM output file")
    parser.add_option("-O", "--offset", dest="offset", help="The alignment offset")
    parser.add_option("-Q", "--minMappingQuality", dest="minMappingQuality", help="The minimum mapping quality")
    parser.add_option(
        "-P",
        "--minAlleleProbability",
        dest="minAlleleProbability",
        help="The minimum allele probability conditioned on coverage (for the binomial quantile).",
    )
    parser.add_option(
        "-C", "--minAlleleCoverage", dest="minAlleleCoverage", help="The minimum haploid coverage for the consensus"
    )
    parser.add_option("-R", "--range", dest="range", help="A range to examine")
    parser.add_option("-c", "--correctBases", dest="correctBases", help="Correct bases ")
    parser.add_option("-q", "--useSequenceQualities", dest="useSequenceQualities", help="Use sequence qualities ")
    parser.add_option(
        "-M",
        "--maxHeapSize",
        dest="maxHeapSize",
        help="The maximum number of nodes on the heap before re-alignment is ignored",
    )
    parser.add_option(
        "-s",
        "--fileSource",
        dest="fileSource",
        help="Whether to use a previously indexed reference sequence or one from history (indexed or history)",
    )
    parser.add_option("-p", "--params", dest="params", help="Parameter setting to use (pre_set or full)")
    parser.add_option("-j", "--jarBin", dest="jarBin", default="", help="The path to where jars are stored")
    parser.add_option("-f", "--jarFile", dest="jarFile", help="The file name of the jar file to use")
    (options, args) = parser.parse_args()

    # make temp directory for srma
    tmp_dir = tempfile.mkdtemp()
    buffsize = 1048576

    # set up reference filenames
    reference_filepath_name = None
    # need to create SRMA dict and Samtools fai files for custom genome
    if options.fileSource == "history":
        try:
            reference_filepath = tempfile.NamedTemporaryFile(dir=tmp_dir, suffix=".fa")
            reference_filepath_name = reference_filepath.name
            reference_filepath.close()
            dict_filepath_name = reference_filepath_name.replace(".fa", ".dict")
            os.symlink(options.ref, reference_filepath_name)
            # create fai file using Samtools
            index_fai_cmd = "samtools faidx %s" % reference_filepath_name
            try:
                tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                tmp_stderr = open(tmp, "wb")
                proc = subprocess.Popen(args=index_fai_cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                returncode = proc.wait()
                tmp_stderr.close()
                # get stderr, allowing for case where it's very large
                tmp_stderr = open(tmp, "rb")
                stderr = ""
                try:
                    while True:
                        stderr += tmp_stderr.read(buffsize)
                        if not stderr or len(stderr) % buffsize != 0:
                            break
                except OverflowError:
                    pass
                tmp_stderr.close()
                if returncode != 0:
                    raise Exception(stderr)
            except Exception as e:
                # clean up temp dir
                if os.path.exists(tmp_dir):
                    shutil.rmtree(tmp_dir)
                stop_err("Error creating Samtools index for custom genome file: %s\n" % str(e))
            # create dict file using SRMA
            dict_cmd = 'java -cp "%s" net.sf.picard.sam.CreateSequenceDictionary R=%s O=%s' % (
                os.path.join(options.jarBin, options.jarFile),
                reference_filepath_name,
                dict_filepath_name,
            )
            try:
                tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                tmp_stderr = open(tmp, "wb")
                proc = subprocess.Popen(args=dict_cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                returncode = proc.wait()
                tmp_stderr.close()
                # get stderr, allowing for case where it's very large
                tmp_stderr = open(tmp, "rb")
                stderr = ""
                try:
                    while True:
                        stderr += tmp_stderr.read(buffsize)
                        if not stderr or len(stderr) % buffsize != 0:
                            break
                except OverflowError:
                    pass
                tmp_stderr.close()
                if returncode != 0:
                    raise Exception(stderr)
            except Exception as e:
                # clean up temp dir
                if os.path.exists(tmp_dir):
                    shutil.rmtree(tmp_dir)
                stop_err("Error creating index for custom genome file: %s\n" % str(e))
        except Exception as e:
            # clean up temp dir
            if os.path.exists(tmp_dir):
                shutil.rmtree(tmp_dir)
            stop_err("Problem handling SRMA index (dict file) for custom genome file: %s\n" % str(e))
    # using built-in dict/index files
    else:
        if options.ref:
            reference_filepath_name = options.ref
        else:
            reference_filepath_name = parseRefLoc(options.refLocation, options.refUID)
    if reference_filepath_name is None:
        raise ValueError("A valid genome reference was not provided.")

    # set up aligning and generate aligning command options
    if options.params == "pre_set":
        srma_cmds = ""
    else:
        ranges = "null"
        if options.range == "None":
            range = "null"
        else:
            range = options.range
        srma_cmds = (
            "OFFSET=%s MIN_MAPQ=%s MINIMUM_ALLELE_PROBABILITY=%s MINIMUM_ALLELE_COVERAGE=%s RANGES=%s RANGE=%s CORRECT_BASES=%s USE_SEQUENCE_QUALITIES=%s MAX_HEAP_SIZE=%s"
            % (
                options.offset,
                options.minMappingQuality,
                options.minAlleleProbability,
                options.minAlleleCoverage,
                ranges,
                range,
                options.correctBases,
                options.useSequenceQualities,
                options.maxHeapSize,
            )
        )

    srma_cmds = "%s VALIDATION_STRINGENCY=LENIENT" % srma_cmds

    # perform alignments
    buffsize = 1048576
    try:
        # symlink input bam and index files due to the naming conventions required by srma here
        input_bam_filename = os.path.join(tmp_dir, "%s.bam" % os.path.split(options.input)[-1])
        os.symlink(options.input, input_bam_filename)
        input_bai_filename = "%s.bai" % os.path.splitext(input_bam_filename)[0]
        os.symlink(options.inputIndex, input_bai_filename)

        # create a temp output name, ending in .bam due to required naming conventions? unkown if required
        output_bam_filename = os.path.join(tmp_dir, "%s.bam" % os.path.split(options.output)[-1])
        # generate commandline
        java_opts = ""
        if "_JAVA_OPTIONS" not in os.environ:
            java_opts = "-Xmx2048m"
        cmd = "java %s -jar %s I=%s O=%s R=%s %s" % (
            java_opts,
            os.path.join(options.jarBin, options.jarFile),
            input_bam_filename,
            output_bam_filename,
            reference_filepath_name,
            srma_cmds,
        )
        # need to nest try-except in try-finally to handle 2.4
        try:
            try:
                tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                tmp_stderr = open(tmp, "wb")
                proc = subprocess.Popen(args=cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                returncode = proc.wait()
                tmp_stderr.close()
                # get stderr, allowing for case where it's very large
                tmp_stderr = open(tmp, "rb")
                stderr = ""
                try:
                    while True:
                        stderr += tmp_stderr.read(buffsize)
                        if not stderr or len(stderr) % buffsize != 0:
                            break
                except OverflowError:
                    pass
                tmp_stderr.close()
                if returncode != 0:
                    raise Exception(stderr)
            except Exception as e:
                raise Exception("Error executing SRMA. " + str(e))
            # move file from temp location (with .bam name) to provided path
            shutil.move(output_bam_filename, options.output)
            # check that there are results in the output file
            if os.path.getsize(options.output) <= 0:
                raise Exception(
                    "The output file is empty. You may simply have no matches, or there may be an error with your input file or settings."
                )
        except Exception as e:
            stop_err("The re-alignment failed.\n" + str(e))
    finally:
        # clean up temp dir
        if os.path.exists(tmp_dir):
            shutil.rmtree(tmp_dir)


if __name__ == "__main__":
    __main__()



""" This module contains functionality to aid in extracting workflows from
histories.
"""

import logging
from typing import Optional

from galaxy import (
    exceptions,
    model,
)
from galaxy.model.base import (
    ensure_object_added_to_session,
    transaction,
)
from galaxy.tool_util.parser import ToolOutputCollectionPart
from galaxy.tools.parameters.basic import (
    DataCollectionToolParameter,
    DataToolParameter,
)
from galaxy.tools.parameters.grouping import (
    Conditional,
    Repeat,
    Section,
)
from galaxy.util import listify
from .steps import (
    attach_ordered_steps,
    order_workflow_steps_with_levels,
)

log = logging.getLogger(__name__)

WARNING_SOME_DATASETS_NOT_READY = "Some datasets still queued or running were ignored"


def extract_workflow(
    trans,
    user,
    history=None,
    job_ids=None,
    dataset_ids=None,
    dataset_collection_ids=None,
    workflow_name=None,
    dataset_names=None,
    dataset_collection_names=None,
):
    steps = extract_steps(
        trans,
        history=history,
        job_ids=job_ids,
        dataset_ids=dataset_ids,
        dataset_collection_ids=dataset_collection_ids,
        dataset_names=dataset_names,
        dataset_collection_names=None,
    )
    # Workflow to populate
    workflow = model.Workflow()
    workflow.name = workflow_name
    workflow.steps = steps
    # Order the steps if possible
    attach_ordered_steps(workflow)
    # And let's try to set up some reasonable locations on the canvas
    # (these are pretty arbitrary values)
    levorder = order_workflow_steps_with_levels(steps)
    base_pos = 10
    for i, steps_at_level in enumerate(levorder):
        for j, index in enumerate(steps_at_level):
            step = steps[index]
            step.position = dict(top=(base_pos + 120 * j), left=(base_pos + 220 * i))
    # Store it
    stored = model.StoredWorkflow()
    stored.user = user
    stored.name = workflow_name
    workflow.stored_workflow = stored
    stored.latest_workflow = workflow
    trans.sa_session.add(stored)
    ensure_object_added_to_session(workflow, session=trans.sa_session)
    with transaction(trans.sa_session):
        trans.sa_session.commit()
    return stored


def extract_steps(
    trans,
    history=None,
    job_ids=None,
    dataset_ids=None,
    dataset_collection_ids=None,
    dataset_names=None,
    dataset_collection_names=None,
):
    # Ensure job_ids and dataset_ids are lists (possibly empty)
    job_ids = listify(job_ids)
    dataset_ids = listify(dataset_ids)
    dataset_collection_ids = listify(dataset_collection_ids)
    # Convert both sets of ids to integers
    job_ids = [int(_) for _ in job_ids]
    dataset_ids = [int(_) for _ in dataset_ids]
    dataset_collection_ids = [int(_) for _ in dataset_collection_ids]
    # Find each job, for security we (implicitly) check that they are
    # associated with a job in the current history.
    summary = WorkflowSummary(trans, history)
    jobs = summary.jobs
    steps = []
    step_labels = set()
    hid_to_output_pair = {}
    # Input dataset steps
    for i, hid in enumerate(dataset_ids):
        step = model.WorkflowStep()
        step.type = "data_input"
        if dataset_names:
            name = dataset_names[i]
        else:
            name = "Input Dataset"
        if name not in step_labels:
            step.label = name
            step_labels.add(name)
        step.tool_inputs = dict(name=name)  # type:ignore[assignment]
        hid_to_output_pair[hid] = (step, "output")
        steps.append(step)
    for i, hid in enumerate(dataset_collection_ids):
        step = model.WorkflowStep()
        step.type = "data_collection_input"
        if hid not in summary.collection_types:
            raise exceptions.RequestParameterInvalidException(f"hid {hid} does not appear to be a collection")
        collection_type = summary.collection_types[hid]
        if dataset_collection_names:
            name = dataset_collection_names[i]
        else:
            name = "Input Dataset Collection"
        if name not in step_labels:
            step.label = name
            step_labels.add(name)
        step.tool_inputs = dict(name=name, collection_type=collection_type)  # type:ignore[assignment]
        hid_to_output_pair[hid] = (step, "output")
        steps.append(step)
    # Tool steps
    for job_id in job_ids:
        if job_id not in summary.job_id2representative_job:
            log.warning(f"job_id {job_id} not found in job_id2representative_job {summary.job_id2representative_job}")
            raise AssertionError("Attempt to create workflow with job not connected to current history")
        job = summary.job_id2representative_job[job_id]
        tool_inputs, associations = step_inputs(trans, job)
        step = model.WorkflowStep()
        step.type = "tool"
        step.tool_id = job.tool_id
        step.tool_version = job.tool_version
        step.tool_inputs = tool_inputs
        # NOTE: We shouldn't need to do two passes here since only
        #       an earlier job can be used as an input to a later
        #       job.
        for other_hid, input_name in associations:
            if job in summary.implicit_map_jobs:
                an_implicit_output_collection = jobs[job][0][1]
                input_collection = an_implicit_output_collection.find_implicit_input_collection(input_name)
                if input_collection:
                    other_hid = input_collection.hid
                else:
                    log.info(f"Cannot find implicit input collection for {input_name}")
            if other_hid in hid_to_output_pair:
                step_input = step.get_or_add_input(input_name)
                other_step, other_name = hid_to_output_pair[other_hid]
                conn = model.WorkflowStepConnection()
                conn.input_step_input = step_input
                # Should always be connected to an earlier step
                conn.output_step = other_step
                conn.output_name = other_name
        steps.append(step)
        # Store created dataset hids
        for assoc in job.output_datasets + job.output_dataset_collection_instances:
            assoc_name = assoc.name
            if ToolOutputCollectionPart.is_named_collection_part_name(assoc_name):
                continue
            if assoc_name.startswith("__new_primary_file"):
                continue
            if job in summary.implicit_map_jobs:
                hid = None
                for implicit_pair in jobs[job]:
                    query_assoc_name, dataset_collection = implicit_pair
                    if query_assoc_name == assoc_name or assoc_name.startswith(
                        f"__new_primary_file_{query_assoc_name}|"
                    ):
                        hid = summary.hid(dataset_collection)
                if hid is None:
                    template = (
                        "Failed to find matching implicit job - job id is %s, implicit pairs are %s, assoc_name is %s."
                    )
                    message = template % (job.id, jobs[job], assoc_name)
                    log.warning(message)
                    raise Exception("Failed to extract job.")
            else:
                if hasattr(assoc, "dataset"):
                    has_hid = assoc.dataset
                else:
                    has_hid = assoc.dataset_collection_instance
                hid = summary.hid(has_hid)
            if hid in hid_to_output_pair:
                log.warning(f"duplicate hid found in extract_steps [{hid}]")
            hid_to_output_pair[hid] = (step, assoc.name)
    return steps


class FakeJob:
    """
    Fake job object for datasets that have no creating_job_associations,
    they will be treated as "input" datasets.
    """

    def __init__(self, dataset):
        self.is_fake = True
        self.id = f"fake_{dataset.id}"
        self.name = self._guess_name_from_dataset(dataset)

    def _guess_name_from_dataset(self, dataset) -> Optional[str]:
        """Tries to guess the name of the fake job from the dataset associations."""
        if dataset.copied_from_history_dataset_association:
            return "Import from History"
        if dataset.copied_from_library_dataset_dataset_association:
            return "Import from Library"
        return None


class DatasetCollectionCreationJob:
    def __init__(self, dataset_collection):
        self.is_fake = True
        self.id = f"fake_{dataset_collection.id}"
        self.from_jobs = None
        self.name = "Dataset Collection Creation"
        self.disabled_why = "Dataset collection created in a way not compatible with workflows"

    def set_jobs(self, jobs):
        assert jobs is not None
        self.from_jobs = jobs


def summarize(trans, history=None):
    """Return mapping of job description to datasets for active items in
    supplied history - needed for building workflow from a history.

    Formerly call get_job_dict in workflow web controller.
    """
    summary = WorkflowSummary(trans, history)
    return summary.jobs, summary.warnings


class WorkflowSummary:
    def __init__(self, trans, history):
        if not history:
            history = trans.get_history()
        self.history = history
        self.warnings = set()
        self.jobs = {}
        self.job_id2representative_job = {}  # map a non-fake job id to its representative job
        self.implicit_map_jobs = []
        self.collection_types = {}

        self.hda_hid_in_history = {}
        self.hdca_hid_in_history = {}

        self.__summarize()

    def hid(self, object):
        if object.history_content_type == "dataset_collection":
            if object.id in self.hdca_hid_in_history:
                return self.hdca_hid_in_history[object.id]
            elif object.history == self.history:
                return object.hid
            else:
                log.warning("extraction issue, using hdca hid from outside current history and unmapped")
                return object.hid
        else:
            if object.id in self.hda_hid_in_history:
                return self.hda_hid_in_history[object.id]
            elif object.history == self.history:
                return object.hid
            else:
                log.warning("extraction issue, using hda hid from outside current history and unmapped")
                return object.hid

    def __summarize(self):
        # Make a first pass handle all singleton jobs, input dataset and dataset collections
        # just grab the implicitly mapped jobs and handle in second pass. Second pass is
        # needed because cannot allow selection of individual datasets from an implicit
        # mapping during extraction - you get the collection or nothing.
        for content in self.history.active_contents:
            self.__summarize_content(content)

    def __summarize_content(self, content):
        # Update internal state for history content (either an HDA or
        # an HDCA).
        if content.history_content_type == "dataset_collection":
            self.__summarize_dataset_collection(content)
        else:
            self.__summarize_dataset(content)

    def __summarize_dataset_collection(self, dataset_collection):
        hid_in_history = dataset_collection.hid
        dataset_collection = self.__original_hdca(dataset_collection)
        self.hdca_hid_in_history[dataset_collection.id] = hid_in_history

        hid = dataset_collection.hid
        self.collection_types[hid] = dataset_collection.collection.collection_type
        if cja := dataset_collection.creating_job_associations:
            # Use the "first" job to represent all mapped jobs.
            representative_assoc = cja[0]
            representative_job = representative_assoc.job
            if (
                representative_job not in self.jobs
                or self.jobs[representative_job][0][1].history_content_type == "dataset"
            ):
                self.jobs[representative_job] = [(representative_assoc.name, dataset_collection)]
                if dataset_collection.implicit_output_name:
                    self.implicit_map_jobs.append(representative_job)
            else:
                self.jobs[representative_job].append((representative_assoc.name, dataset_collection))
            for assoc in cja:
                job = assoc.job
                self.job_id2representative_job[job.id] = representative_job
        # This whole elif condition may no longer be needed do to additional
        # tracking with creating_job_associations. Will delete at some point.
        elif dataset_collection.implicit_output_name:
            # TODO: Optimize db call
            element = dataset_collection.collection.first_dataset_element
            if not element:
                # Got no dataset instance to walk back up to creating job.
                # TODO track this via tool request model
                job = DatasetCollectionCreationJob(dataset_collection)
                self.jobs[job] = [(None, dataset_collection)]
                return
            else:
                dataset_instance = element.hda
            if not self.__check_state(dataset_instance):
                # Just checking the state of one instance, don't need more but
                # makes me wonder if even need this check at all?
                return

            original_hda = self.__original_hda(dataset_instance)
            if not original_hda.creating_job_associations:
                log.warning(
                    "An implicitly create output dataset collection doesn't have a creating_job_association, should not happen!"
                )
                job = DatasetCollectionCreationJob(dataset_collection)
                self.jobs[job] = [(None, dataset_collection)]

            for assoc in original_hda.creating_job_associations:
                job = assoc.job
                if job not in self.jobs or self.jobs[job][0][1].history_content_type == "dataset":
                    self.jobs[job] = [(assoc.name, dataset_collection)]
                    self.job_id2representative_job[job.id] = job
                    self.implicit_map_jobs.append(job)
                else:
                    self.jobs[job].append((assoc.name, dataset_collection))
        else:
            job = DatasetCollectionCreationJob(dataset_collection)
            self.jobs[job] = [(None, dataset_collection)]

    def __summarize_dataset(self, dataset):
        if not self.__check_state(dataset):
            return

        hid_in_history = dataset.hid
        original_hda = self.__original_hda(dataset)
        self.hda_hid_in_history[original_hda.id] = hid_in_history

        if not original_hda.creating_job_associations:
            self.jobs[FakeJob(dataset)] = [(None, dataset)]

        for assoc in original_hda.creating_job_associations:
            job = assoc.job
            if job in self.jobs:
                self.jobs[job].append((assoc.name, dataset))
            else:
                self.jobs[job] = [(assoc.name, dataset)]
                self.job_id2representative_job[job.id] = job

    def __original_hdca(self, hdca):
        while hdca.copied_from_history_dataset_collection_association:
            hdca = hdca.copied_from_history_dataset_collection_association
        return hdca

    def __original_hda(self, hda):
        # if this hda was copied from another, we need to find the job that created the original hda
        while hda.copied_from_history_dataset_association:
            hda = hda.copied_from_history_dataset_association
        return hda

    def __check_state(self, hda):
        # FIXME: Create "Dataset.is_finished"
        if hda.state in ("new", "running", "queued"):
            self.warnings.add(WARNING_SOME_DATASETS_NOT_READY)
            return
        return hda


def step_inputs(trans, job):
    tool = trans.app.toolbox.get_tool(job.tool_id, tool_version=job.tool_version)
    param_values = job.get_param_values(
        trans.app, ignore_errors=True
    )  # If a tool was updated and e.g. had a text value changed to an integer, we don't want a traceback here
    associations = __cleanup_param_values(tool.inputs, param_values)
    tool_inputs = tool.params_to_strings(param_values, trans.app)
    return tool_inputs, associations


def __cleanup_param_values(inputs, values):
    """
    Remove 'Data' values from `param_values`, along with metadata cruft,
    but track the associations.
    """
    associations = []
    # dbkey is pushed in by the framework
    if "dbkey" in values:
        del values["dbkey"]
    root_values = values
    root_input_keys = inputs.keys()

    # Recursively clean data inputs and dynamic selects
    def cleanup(prefix, inputs, values):
        for key, input in inputs.items():
            if isinstance(input, DataToolParameter) or isinstance(input, DataCollectionToolParameter):
                items = values[key]
                values[key] = None
                # HACK: Nested associations are not yet working, but we
                #       still need to clean them up so we can serialize
                # if not( prefix ):
                for item in listify(items):
                    if isinstance(item, model.DatasetCollectionElement):
                        item = item.first_dataset_instance()
                    if item:  # this is false for a non-set optional dataset
                        associations.append((item.hid, prefix + key))

                # Cleanup the other deprecated crap associated with datasets
                # as well. Worse, for nested datasets all the metadata is
                # being pushed into the root. FIXME: MUST REMOVE SOON
                key = f"{prefix + key}_"
                for k in root_values.keys():
                    if k not in root_input_keys and k.startswith(key):
                        del root_values[k]
            elif isinstance(input, Repeat):
                if key in values:
                    group_values = values[key]
                    for i, rep_values in enumerate(group_values):
                        rep_index = rep_values["__index__"]
                        cleanup("%s%s_%d|" % (prefix, key, rep_index), input.inputs, group_values[i])
            elif isinstance(input, Conditional):
                # Scrub dynamic resource related parameters from workflows,
                # they cause problems and the workflow probably should include
                # their state in workflow encoding.
                if input.name == "__job_resource":
                    if input.name in values:
                        del values[input.name]
                    return
                if input.name in values:
                    group_values = values[input.name]
                    current_case = group_values["__current_case__"]
                    cleanup(f"{prefix}{key}|", input.cases[current_case].inputs, group_values)
            elif isinstance(input, Section):
                if input.name in values:
                    cleanup(f"{prefix}{key}|", input.inputs, values[input.name])

    cleanup("", inputs, values)
    return associations


__all__ = ("summarize", "extract_workflow")



"""
Classes related to parameter validation.
"""

import abc
import json
import logging
import os.path

import regex

from galaxy import (
    model,
    util,
)

log = logging.getLogger(__name__)


class Validator(abc.ABC):
    """
    A validator checks that a value meets some conditions OR raises ValueError
    """

    requires_dataset_metadata = False

    @classmethod
    def from_element(cls, param, elem):
        """
        Initialize the appropriate Validator class

        example call `validation.Validator.from_element(ToolParameter_object, Validator_object)`

        needs to be implemented in the subclasses and should return the
        corresponding Validator object by a call to `cls( ... )` which calls the
        `__init__` method of the corresponding validator

        param cls the Validator class
        param param the element to be evaluated (which contains the validator)
        param elem the validator element
        return an object of a Validator subclass that corresponds to the type attribute of the validator element
        """
        _type = elem.get("type")
        assert _type is not None, "Required 'type' attribute missing from validator"
        return validator_types[_type].from_element(param, elem)

    def __init__(self, message, negate=False):
        self.message = message
        self.negate = util.asbool(negate)
        super().__init__()

    @abc.abstractmethod
    def validate(self, value, trans=None, message=None, value_to_show=None):
        """
        validate a value

        needs to be implemented in classes derived from validator.
        the implementation needs to call `super().validate()`
        giving result as a bool (which should be true if the
        validation is positive and false otherwise) and the value
        that is validated.

        the Validator.validate function will then negate the value
        depending on `self.negate` and return None if
        - value is True and negate is False
        - value is False and negate is True
        and raise a ValueError otherwise.

        return None if positive validation, otherwise a ValueError is raised
        """
        assert isinstance(value, bool), "value must be boolean"
        if message is None:
            message = self.message
        if value_to_show and "%s" in message:
            message = message % value_to_show
        if (not self.negate and value) or (self.negate and not value):
            return
        else:
            raise ValueError(message)


class RegexValidator(Validator):
    """
    Validator that evaluates a regular expression
    """

    @classmethod
    def from_element(cls, param, elem):
        return cls(elem.get("message"), elem.text, elem.get("negate", "false"))

    def __init__(self, message, expression, negate):
        if message is None:
            message = f"Value '%s' does {'not ' if negate == 'false' else ''}match regular expression '{expression.replace('%', '%%')}'"
        super().__init__(message, negate)
        # Compile later. RE objects used to not be thread safe. Not sure about
        # the sre module.
        self.expression = expression

    def validate(self, value, trans=None):
        if not isinstance(value, list):
            value = [value]
        for val in value:
            match = regex.match(self.expression, val or "")
            super().validate(match is not None, value_to_show=val)


class ExpressionValidator(Validator):
    """
    Validator that evaluates a python expression using the value
    """

    @classmethod
    def from_element(cls, param, elem):
        return cls(elem.get("message"), elem.text, elem.get("negate", "false"))

    def __init__(self, message, expression, negate):
        if message is None:
            message = f"Value '%s' does not evaluate to {'True' if negate == 'false' else 'False'} for '{expression}'"
        super().__init__(message, negate)
        self.expression = expression
        # Save compiled expression, code objects are thread safe (right?)
        self.compiled_expression = compile(expression, "<string>", "eval")

    def validate(self, value, trans=None):
        try:
            evalresult = eval(self.compiled_expression, dict(value=value))
        except Exception:
            super().validate(False, message=f"Validator '{self.expression}' could not be evaluated on '{value}'")
        super().validate(bool(evalresult), value_to_show=value)


class InRangeValidator(ExpressionValidator):
    """
    Validator that ensures a number is in a specified range
    """

    @classmethod
    def from_element(cls, param, elem):
        return cls(
            elem.get("message"),
            elem.get("min"),
            elem.get("max"),
            elem.get("exclude_min", "false"),
            elem.get("exclude_max", "false"),
            elem.get("negate", "false"),
        )

    def __init__(self, message, range_min, range_max, exclude_min=False, exclude_max=False, negate=False):
        """
        When the optional exclude_min and exclude_max attributes are set
        to true, the range excludes the end points (i.e., min < value < max),
        while if set to False (the default), then range includes the end points
        (1.e., min <= value <= max).  Combinations of exclude_min and exclude_max
        values are allowed.
        """
        self.min = range_min if range_min is not None else "-inf"
        self.exclude_min = util.asbool(exclude_min)
        self.max = range_max if range_max is not None else "inf"
        self.exclude_max = util.asbool(exclude_max)
        assert float(self.min) <= float(self.max), "min must be less than or equal to max"
        # Remove unneeded 0s and decimal from floats to make message pretty.
        op1 = "<="
        op2 = "<="
        if self.exclude_min:
            op1 = "<"
        if self.exclude_max:
            op2 = "<"
        expression = f"float('{self.min}') {op1} float(value) {op2} float('{self.max}')"
        if message is None:
            message = f"Value ('%s') must {'not ' if negate == 'true' else ''}fulfill {expression}"
        super().__init__(message, expression, negate)


class LengthValidator(InRangeValidator):
    """
    Validator that ensures the length of the provided string (value) is in a specific range
    """

    @classmethod
    def from_element(cls, param, elem):
        return cls(elem.get("message"), elem.get("min"), elem.get("max"), elem.get("negate", "false"))

    def __init__(self, message, length_min, length_max, negate):
        if message is None:
            message = f"Must {'not ' if negate == 'true' else ''}have length of at least {length_min} and at most {length_max}"
        super().__init__(message, range_min=length_min, range_max=length_max, negate=negate)

    def validate(self, value, trans=None):
        if value is None:
            raise ValueError("No value provided")
        super().validate(len(value) if value else 0, trans)


class DatasetOkValidator(Validator):
    """
    Validator that checks if a dataset is in an 'ok' state
    """

    @classmethod
    def from_element(cls, param, elem):
        negate = elem.get("negate", "false")
        message = elem.get("message")
        if message is None:
            if negate == "false":
                message = "The selected dataset is still being generated, select another dataset or wait until it is completed"
            else:
                message = "The selected dataset must not be in state OK"
        return cls(message, negate)

    def validate(self, value, trans=None):
        if value:
            super().validate(value.state == model.Dataset.states.OK)


class DatasetEmptyValidator(Validator):
    """
    Validator that checks if a dataset has a positive file size.
    """

    @classmethod
    def from_element(cls, param, elem):
        message = elem.get("message")
        negate = elem.get("negate", "false")
        if not message:
            message = f"The selected dataset is {'non-' if negate == 'true' else ''}empty, this tool expects {'non-' if negate == 'false' else ''}empty files."
        return cls(message, negate)

    def validate(self, value, trans=None):
        if value:
            super().validate(value.get_size() != 0)


class DatasetExtraFilesPathEmptyValidator(Validator):
    """
    Validator that checks if a dataset's extra_files_path exists and is not empty.
    """

    @classmethod
    def from_element(cls, param, elem):
        message = elem.get("message")
        negate = elem.get("negate", "false")
        if not message:
            message = f"The selected dataset's extra_files_path directory is {'non-' if negate == 'true' else ''}empty or does {'not ' if negate == 'false' else ''}exist, this tool expects {'non-' if negate == 'false' else ''}empty extra_files_path directories associated with the selected input."
        return cls(message, negate)

    def validate(self, value, trans=None):
        if value:
            super().validate(value.get_total_size() != value.get_size())


class MetadataValidator(Validator):
    """
    Validator that checks for missing metadata
    """

    requires_dataset_metadata = True

    @classmethod
    def from_element(cls, param, elem):
        message = elem.get("message")
        return cls(
            message=message, check=elem.get("check", ""), skip=elem.get("skip", ""), negate=elem.get("negate", "false")
        )

    def __init__(self, message=None, check="", skip="", negate="false"):
        if not message:
            if not util.asbool(negate):
                message = "Metadata '%s' missing, click the pencil icon in the history item to edit / save the metadata attributes"
            else:
                if check != "":
                    message = f"At least one of the checked metadata '{check}' is set, click the pencil icon in the history item to edit / save the metadata attributes"
                elif skip != "":
                    message = f"At least one of the non skipped metadata '{skip}' is set, click the pencil icon in the history item to edit / save the metadata attributes"
        super().__init__(message, negate)
        self.check = check.split(",") if check else None
        self.skip = skip.split(",") if skip else None

    def validate(self, value, trans=None):
        if value:
            # TODO why this validator checks for isinstance(value, model.DatasetInstance)
            missing = value.missing_meta(check=self.check, skip=self.skip)
            super().validate(isinstance(value, model.DatasetInstance) and not missing, value_to_show=missing)


class MetadataEqualValidator(Validator):
    """
    Validator that checks for a metadata value for equality

    metadata values that are lists are converted as comma separated string
    everything else is converted to the string representation
    """

    requires_dataset_metadata = True

    def __init__(self, metadata_name=None, value=None, message=None, negate="false"):
        if not message:
            if not util.asbool(negate):
                message = f"Metadata value for '{metadata_name}' must be '{value}', but it is '%s'."
            else:
                message = f"Metadata value for '{metadata_name}' must not be '{value}' but it is."
        super().__init__(message, negate)
        self.metadata_name = metadata_name
        self.value = value

    @classmethod
    def from_element(cls, param, elem):
        value = elem.get("value", None) or json.loads(elem.get("value_json", "null"))
        return cls(
            metadata_name=elem.get("metadata_name", None),
            value=value,
            message=elem.get("message", None),
            negate=elem.get("negate", "false"),
        )

    def validate(self, value, trans=None):
        if value:
            metadata_value = getattr(value.metadata, self.metadata_name)
            super().validate(metadata_value == self.value, value_to_show=metadata_value)


class UnspecifiedBuildValidator(Validator):
    """
    Validator that checks for dbkey not equal to '?'
    """

    requires_dataset_metadata = True

    @classmethod
    def from_element(cls, param, elem):
        message = elem.get("message")
        negate = elem.get("negate", "false")
        if not message:
            message = f"{'Unspecified' if negate == 'false' else 'Specified'} genome build, click the pencil icon in the history item to {'set' if negate == 'false' else 'remove'} the genome build"
        return cls(message, negate)

    def validate(self, value, trans=None):
        # if value is None, we cannot validate
        if value:
            dbkey = value.metadata.dbkey
            # TODO can dbkey really be a list?
            if isinstance(dbkey, list):
                dbkey = dbkey[0]
            super().validate(dbkey != "?")


class NoOptionsValidator(Validator):
    """
    Validator that checks for empty select list
    """

    @classmethod
    def from_element(cls, param, elem):
        message = elem.get("message")
        negate = elem.get("negate", "false")
        if not message:
            message = f"{'No options' if negate == 'false' else 'Options'} available for selection"
        return cls(message, negate)

    def validate(self, value, trans=None):
        super().validate(value is not None)


class EmptyTextfieldValidator(Validator):
    """
    Validator that checks for empty text field
    """

    @classmethod
    def from_element(cls, param, elem):
        message = elem.get("message")
        negate = elem.get("negate", "false")
        if not message:
            if negate == "false":
                message = elem.get("message", "Field requires a value")
            else:
                message = elem.get("message", "Field must not set a value")
        return cls(message, negate)

    def validate(self, value, trans=None):
        super().validate(value != "")


class MetadataInFileColumnValidator(Validator):
    """
    Validator that checks if the value for a dataset's metadata item exists in a file.

    Deprecated: DataTables are now the preferred way.

    note: this is covered in a framework test (validation_dataset_metadata_in_file)
    """

    requires_dataset_metadata = True

    @classmethod
    def from_element(cls, param, elem):
        filename = elem.get("filename")
        assert filename, f"Required 'filename' attribute missing from {elem.get('type')} validator."
        filename = f"{param.tool.app.config.tool_data_path}/{filename.strip()}"
        assert os.path.exists(filename), f"File {filename} specified by the 'filename' attribute not found"
        metadata_name = elem.get("metadata_name")
        assert metadata_name, f"Required 'metadata_name' attribute missing from {elem.get('type')} validator."
        metadata_name = metadata_name.strip()
        metadata_column = int(elem.get("metadata_column", 0))
        split = elem.get("split", "\t")
        message = elem.get("message", f"Value for metadata {metadata_name} was not found in {filename}.")
        line_startswith = elem.get("line_startswith")
        if line_startswith:
            line_startswith = line_startswith.strip()
        negate = elem.get("negate", "false")
        return cls(filename, metadata_name, metadata_column, message, line_startswith, split, negate)

    def __init__(
        self,
        filename,
        metadata_name,
        metadata_column,
        message="Value for metadata not found.",
        line_startswith=None,
        split="\t",
        negate="false",
    ):
        super().__init__(message, negate)
        self.metadata_name = metadata_name
        self.valid_values = set()
        with open(filename) as fh:
            for line in fh:
                if line_startswith is None or line.startswith(line_startswith):
                    fields = line.split(split)
                    if metadata_column < len(fields):
                        self.valid_values.add(fields[metadata_column].strip())

    def validate(self, value, trans=None):
        if not value:
            return
        super().validate(
            value.metadata.spec[self.metadata_name].param.to_string(value.metadata.get(self.metadata_name))
            in self.valid_values
        )


class ValueInDataTableColumnValidator(Validator):
    """
    Validator that checks if a value is in a tool data table column.

    note: this is covered in a framework test (validation_value_in_datatable)
    """

    @classmethod
    def from_element(cls, param, elem):
        table_name = elem.get("table_name")
        assert table_name, f"Required 'table_name' attribute missing from {elem.get('type')} validator."
        tool_data_table = param.tool.app.tool_data_tables[table_name]
        column = elem.get("metadata_column", 0)
        try:
            column = int(column)
        except ValueError:
            pass
        message = elem.get("message", f"Value was not found in {table_name}.")
        negate = elem.get("negate", "false")
        return cls(tool_data_table, column, message, negate)

    def __init__(self, tool_data_table, column, message="Value not found.", negate="false"):
        super().__init__(message, negate)
        self.valid_values = []
        self._data_table_content_version = None
        self._tool_data_table = tool_data_table
        if isinstance(column, str):
            column = tool_data_table.columns[column]
        self._column = column
        self._load_values()

    def _load_values(self):
        self._data_table_content_version, data_fields = self._tool_data_table.get_version_fields()
        self.valid_values = []
        for fields in data_fields:
            if self._column < len(fields):
                self.valid_values.append(fields[self._column])

    def validate(self, value, trans=None):
        if not value:
            return
        if not self._tool_data_table.is_current_version(self._data_table_content_version):
            log.debug(
                "ValueInDataTableColumnValidator: values are out of sync with data table (%s), updating validator.",
                self._tool_data_table.name,
            )
            self._load_values()
        super().validate(value in self.valid_values)


class ValueNotInDataTableColumnValidator(ValueInDataTableColumnValidator):
    """
    Validator that checks if a value is NOT in a tool data table column.
    Equivalent to ValueInDataTableColumnValidator with `negate="true"`.

    note: this is covered in a framework test (validation_value_in_datatable)
    """

    def __init__(self, tool_data_table, metadata_column, message="Value already present.", negate="false"):
        super().__init__(tool_data_table, metadata_column, message, negate)

    def validate(self, value, trans=None):
        try:
            super().validate(value)
        except ValueError:
            return
        else:
            raise ValueError(self.message)


class MetadataInDataTableColumnValidator(ValueInDataTableColumnValidator):
    """
    Validator that checks if the value for a dataset's metadata item exists in a file.

    note: this is covered in a framework test (validation_metadata_in_datatable)
    """

    requires_dataset_metadata = True

    @classmethod
    def from_element(cls, param, elem):
        table_name = elem.get("table_name")
        assert table_name, f"Required 'table_name' attribute missing from {elem.get('type')} validator."
        tool_data_table = param.tool.app.tool_data_tables[table_name]
        metadata_name = elem.get("metadata_name")
        assert metadata_name, f"Required 'metadata_name' attribute missing from {elem.get('type')} validator."
        metadata_name = metadata_name.strip()
        # TODO rename to column?
        metadata_column = elem.get("metadata_column", 0)
        try:
            metadata_column = int(metadata_column)
        except ValueError:
            pass
        message = elem.get("message", f"Value for metadata {metadata_name} was not found in {table_name}.")
        negate = elem.get("negate", "false")
        return cls(tool_data_table, metadata_name, metadata_column, message, negate)

    def __init__(
        self, tool_data_table, metadata_name, metadata_column, message="Value for metadata not found.", negate="false"
    ):
        super().__init__(tool_data_table, metadata_column, message, negate)
        self.metadata_name = metadata_name

    def validate(self, value, trans=None):
        super().validate(
            value.metadata.spec[self.metadata_name].param.to_string(value.metadata.get(self.metadata_name)), trans
        )


class MetadataNotInDataTableColumnValidator(MetadataInDataTableColumnValidator):
    """
    Validator that checks if the value for a dataset's metadata item doesn't exists in a file.
    Equivalent to MetadataInDataTableColumnValidator with `negate="true"`.

    note: this is covered in a framework test (validation_metadata_in_datatable)
    """

    requires_dataset_metadata = True

    def __init__(
        self, tool_data_table, metadata_name, metadata_column, message="Value for metadata not found.", negate="false"
    ):
        super().__init__(tool_data_table, metadata_name, metadata_column, message, negate)

    def validate(self, value, trans=None):
        try:
            super().validate(value, trans)
        except ValueError:
            return
        else:
            raise ValueError(self.message)


class MetadataInRangeValidator(InRangeValidator):
    """
    validator that ensures metadata is in a specified range

    note: this is covered in a framework test (validation_metadata_in_range)
    """

    requires_dataset_metadata = True

    @classmethod
    def from_element(cls, param, elem):
        metadata_name = elem.get("metadata_name")
        assert metadata_name, f"Required 'metadata_name' attribute missing from {elem.get('type')} validator."
        metadata_name = metadata_name.strip()
        ret = cls(
            metadata_name,
            elem.get("message"),
            elem.get("min"),
            elem.get("max"),
            elem.get("exclude_min", "false"),
            elem.get("exclude_max", "false"),
            elem.get("negate", "false"),
        )
        ret.message = "Metadata: " + ret.message
        return ret

    def __init__(self, metadata_name, message, range_min, range_max, exclude_min, exclude_max, negate):
        self.metadata_name = metadata_name
        super().__init__(message, range_min, range_max, exclude_min, exclude_max, negate)

    def validate(self, value, trans=None):
        if value:
            if not isinstance(value, model.DatasetInstance):
                raise ValueError("A non-dataset value was provided.")
            try:
                value_to_check = float(
                    value.metadata.spec[self.metadata_name].param.to_string(value.metadata.get(self.metadata_name))
                )
            except KeyError:
                raise ValueError(f"{self.metadata_name} Metadata missing")
            except ValueError:
                raise ValueError(f"{self.metadata_name} must be a float or an integer")
            super().validate(value_to_check, trans)


validator_types = dict(
    expression=ExpressionValidator,
    regex=RegexValidator,
    in_range=InRangeValidator,
    length=LengthValidator,
    metadata=MetadataValidator,
    dataset_metadata_equal=MetadataEqualValidator,
    unspecified_build=UnspecifiedBuildValidator,
    no_options=NoOptionsValidator,
    empty_field=EmptyTextfieldValidator,
    empty_dataset=DatasetEmptyValidator,
    empty_extra_files_path=DatasetExtraFilesPathEmptyValidator,
    dataset_metadata_in_data_table=MetadataInDataTableColumnValidator,
    dataset_metadata_not_in_data_table=MetadataNotInDataTableColumnValidator,
    dataset_metadata_in_range=MetadataInRangeValidator,
    value_in_data_table=ValueInDataTableColumnValidator,
    value_not_in_data_table=ValueNotInDataTableColumnValidator,
    dataset_ok_validator=DatasetOkValidator,
)

deprecated_validator_types = dict(dataset_metadata_in_file=MetadataInFileColumnValidator)
validator_types.update(deprecated_validator_types)



def exec_before_job(app, inp_data, out_data, param_dict, tool):
    out_data["output1"].name = out_data["output1"].name + " [ ALIGNMENT INFO ]"
    out_data["output2"].name = out_data["output2"].name + " [ PILEUP ]"
    out_data["output3"].name = out_data["output3"].name + " [ CUSTOM TRACK ]"



import logging
from collections import defaultdict
from typing import (
    Dict,
    List,
    Tuple,
)

from sqlalchemy.orm import defer

from galaxy.model.scoped_session import install_model_scoped_session
from galaxy.model.tool_shed_install import ToolShedRepository
from galaxy.tool_util.toolbox.base import ToolConfRepository

log = logging.getLogger(__name__)


class ToolShedRepositoryCache:
    """
    Cache installed ToolShedRepository objects.
    """

    local_repositories: List[ToolConfRepository]
    repositories: List[ToolShedRepository]
    repos_by_tuple: Dict[Tuple[str, str, str], List[ToolConfRepository]]

    def __init__(self, session: install_model_scoped_session):
        self.session = session()
        # Contains ToolConfRepository objects created from shed_tool_conf.xml entries
        self.local_repositories = []
        # Repositories loaded from database
        self.repositories = []
        self.repos_by_tuple = defaultdict(list)
        self._build()
        self.session.close()

    def add_local_repository(self, repository):
        self.local_repositories.append(repository)
        self.repos_by_tuple[(repository.tool_shed, repository.owner, repository.name)].append(repository)

    def _build(self):
        self.repositories = self.session.query(ToolShedRepository).options(defer(ToolShedRepository.metadata_)).all()
        repos_by_tuple = defaultdict(list)
        for repository in self.repositories + self.local_repositories:
            repos_by_tuple[(repository.tool_shed, repository.owner, repository.name)].append(repository)
        self.repos_by_tuple = repos_by_tuple

    def get_installed_repository(
        self,
        tool_shed=None,
        name=None,
        owner=None,
        installed_changeset_revision=None,
        changeset_revision=None,
        repository_id=None,
    ):
        if repository_id:
            repos = [repo for repo in self.repositories if repo.id == repository_id]
            if repos:
                return repos[0]
            else:
                return None
        repos = self.repos_by_tuple[(tool_shed, owner, name)]
        for repo in repos:
            if installed_changeset_revision and repo.installed_changeset_revision != installed_changeset_revision:
                continue
            if changeset_revision and repo.changeset_revision != changeset_revision:
                continue
            return repo
        return None



import errno
import logging
import os
import time
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from galaxy.tool_shed.galaxy_install.client import (
    DataManagerInterface,
    InstallationTarget,
)
from galaxy.util import (
    Element,
    etree,
    parse_xml_string,
    xml_to_string,
)
from galaxy.util.path import StrPath
from galaxy.util.renamed_temporary_file import RenamedTemporaryFile
from galaxy.util.tool_shed.xml_util import parse_xml
from . import tool_panel_manager

log = logging.getLogger(__name__)

SHED_DATA_MANAGER_CONF_XML = """<?xml version="1.0"?>
<data_managers>
</data_managers>
"""


class DataManagerHandler:
    app: InstallationTarget
    root: Optional[Element] = None

    def __init__(self, app: InstallationTarget):
        self.app = app

    @property
    def data_managers_path(self) -> Optional[str]:
        tree, error_message = parse_xml(self.app.config.shed_data_manager_config_file)
        if tree:
            root = tree.getroot()
            return root.get("tool_path", None)
        return None

    def _data_manager_config_elems_to_xml_file(self, config_elems: List[Element], config_filename: StrPath) -> None:
        """
        Persist the current in-memory list of config_elems to a file named by the value
        of config_filename.
        """
        if data_managers_path := self.data_managers_path:
            root_str = f'<?xml version="1.0"?><data_managers tool_path="{data_managers_path}"></data_managers>'
        else:
            root_str = '<?xml version="1.0"?><data_managers></data_managers>'
        root = parse_xml_string(root_str)
        for elem in config_elems:
            root.append(elem)
        try:
            with RenamedTemporaryFile(config_filename, mode="w") as fh:
                fh.write(xml_to_string(root))
        except Exception:
            log.exception("Exception in DataManagerHandler._data_manager_config_elems_to_xml_file")

    def install_data_managers(
        self,
        shed_data_manager_conf_filename: StrPath,
        metadata_dict: Dict[str, Any],
        shed_config_dict: Dict[str, Any],
        relative_install_dir: StrPath,
        repository,
        repository_tools_tups,
    ) -> List["DataManagerInterface"]:
        rval: List["DataManagerInterface"] = []
        if "data_manager" in metadata_dict:
            tpm = tool_panel_manager.ToolPanelManager(self.app)
            repository_tools_by_guid = {}
            for tool_tup in repository_tools_tups:
                repository_tools_by_guid[tool_tup[1]] = dict(tool_config_filename=tool_tup[0], tool=tool_tup[2])
            # Load existing data managers.
            try:
                tree, error_message = parse_xml(shed_data_manager_conf_filename, check_exists=False)
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    with open(shed_data_manager_conf_filename, "w") as fh:
                        fh.write(SHED_DATA_MANAGER_CONF_XML)
                    tree, error_message = parse_xml(shed_data_manager_conf_filename)
                else:
                    raise
            if tree is None:
                return rval
            config_elems = list(
                tree.getroot().__iter__()
            )  # `.__iter__()` is a workaround for lxml-stubs declaring _Element a subclass of Iterable["_Element"]
            repo_data_manager_conf_filename = metadata_dict["data_manager"].get("config_filename", None)
            if repo_data_manager_conf_filename is None:
                log.debug("No data_manager_conf.xml file has been defined.")
                return rval
            data_manager_config_has_changes = False
            relative_repo_data_manager_dir = os.path.join(shed_config_dict.get("tool_path", ""), relative_install_dir)
            repo_data_manager_conf_filename = os.path.join(
                relative_repo_data_manager_dir, repo_data_manager_conf_filename
            )
            tree, error_message = parse_xml(repo_data_manager_conf_filename)
            if tree is None:
                return rval
            root = tree.getroot()
            for elem in root:
                if elem.tag == "data_manager":
                    data_manager_id = elem.get("id", None)
                    if data_manager_id is None:
                        log.error(
                            "A data manager was defined that does not have an id and will not be installed:\n%s",
                            xml_to_string(elem),
                        )
                        continue
                    data_manager_dict = (
                        metadata_dict["data_manager"].get("data_managers", {}).get(data_manager_id, None)
                    )
                    if data_manager_dict is None:
                        log.error(f"Data manager metadata is not defined properly for '{data_manager_id}'.")
                        continue
                    guid = data_manager_dict.get("guid", None)
                    if guid is None:
                        log.error(f"Data manager guid '{guid}' is not set in metadata for '{data_manager_id}'.")
                        continue
                    elem.set("guid", guid)
                    tool_guid = data_manager_dict.get("tool_guid", None)
                    if tool_guid is None:
                        log.error(
                            f"Data manager tool guid '{tool_guid}' is not set in metadata for '{data_manager_id}'."
                        )
                        continue
                    tool_dict = repository_tools_by_guid.get(tool_guid, None)
                    if tool_dict is None:
                        log.error(
                            "Data manager tool guid '%s' could not be found for '%s'. Perhaps the tool is invalid?",
                            tool_guid,
                            data_manager_id,
                        )
                        continue
                    tool = tool_dict.get("tool", None)
                    if tool is None:
                        log.error(
                            "Data manager tool with guid '%s' could not be found for '%s'. Perhaps the tool is invalid?",
                            tool_guid,
                            data_manager_id,
                        )
                        continue
                    tool_config_filename = tool_dict.get("tool_config_filename", None)
                    if tool_config_filename is None:
                        log.error(f"Data manager metadata is missing 'tool_config_file' for '{data_manager_id}'.")
                        continue
                    elem.set("shed_conf_file", shed_config_dict["config_filename"])
                    if elem.get("tool_file", None) is not None:
                        del elem.attrib["tool_file"]  # remove old tool_file info
                    tool_elem = tpm.generate_tool_elem(
                        repository.tool_shed,
                        repository.name,
                        repository.installed_changeset_revision,
                        repository.owner,
                        tool_config_filename,
                        tool,
                        None,
                    )
                    elem.insert(0, tool_elem)
                    data_manager = self.app.data_managers.load_manager_from_elem(
                        elem, tool_path=shed_config_dict.get("tool_path", "")
                    )
                    if data_manager:
                        rval.append(data_manager)
                elif elem.tag is etree.Comment:  # type: ignore[comparison-overlap]
                    pass
                else:
                    log.warning(f"Encountered unexpected element '{elem.tag}':\n{xml_to_string(elem)}")
                config_elems.append(elem)
                data_manager_config_has_changes = True
            # Persist the altered shed_data_manager_config file.
            if data_manager_config_has_changes:
                reload_count = self.app.data_managers._reload_count
                self._data_manager_config_elems_to_xml_file(config_elems, shed_data_manager_conf_filename)
                while self.app.data_managers._reload_count <= reload_count:
                    time.sleep(0.1)  # Wait for shed_data_manager watcher thread to pick up changes
        return rval

    def remove_from_data_manager(self, repository):
        metadata_dict = repository.metadata_
        if metadata_dict and "data_manager" in metadata_dict:
            shed_data_manager_conf_filename = self.app.config.shed_data_manager_config_file
            tree, error_message = parse_xml(shed_data_manager_conf_filename)
            if tree:
                root = tree.getroot()
                assert (
                    root.tag == "data_managers"
                ), f"The file provided ({shed_data_manager_conf_filename}) for removing data managers from is not a valid data manager xml file."
                guids = [
                    data_manager_dict.get("guid")
                    for data_manager_dict in metadata_dict.get("data_manager", {}).get("data_managers", {}).values()
                    if "guid" in data_manager_dict
                ]
                load_old_data_managers_by_guid = {}
                data_manager_config_has_changes = False
                config_elems = []
                for elem in root:
                    # Match Data Manager elements by guid and installed_changeset_revision
                    elem_matches_removed_data_manager = False
                    if elem.tag == "data_manager":
                        guid = elem.get("guid", None)
                        if guid in guids:
                            tool_elem = elem.find("tool")
                            if tool_elem is not None:
                                installed_changeset_revision_elem = tool_elem.find("installed_changeset_revision")
                                if installed_changeset_revision_elem is not None:
                                    if (
                                        installed_changeset_revision_elem.text
                                        == repository.installed_changeset_revision
                                    ):
                                        elem_matches_removed_data_manager = True
                                    else:
                                        # This is a different version, which had been previously overridden
                                        load_old_data_managers_by_guid[guid] = elem
                    if elem_matches_removed_data_manager:
                        data_manager_config_has_changes = True
                    else:
                        config_elems.append(elem)
                # Remove data managers from in memory
                self.app.data_managers.remove_manager(guids)
                # Load other versions of any now uninstalled data managers, if any
                for elem in load_old_data_managers_by_guid.values():
                    self.app.data_managers.load_manager_from_elem(elem)
                # Persist the altered shed_data_manager_config file.
                if data_manager_config_has_changes:
                    self._data_manager_config_elems_to_xml_file(config_elems, shed_data_manager_conf_filename)



import logging
import os
import shutil

import markupsafe

from galaxy.util import (
    nice_size,
    unicodify,
)

log = logging.getLogger(__name__)

CHUNK_SIZE = 2**20  # 1Mb
INSTALLATION_LOG = "INSTALLATION.log"
# Set no activity timeout to 20 minutes.
NO_OUTPUT_TIMEOUT = 3600.0
MAXDIFFSIZE = 8000
MAX_DISPLAY_SIZE = 32768

DOCKER_IMAGE_TEMPLATE = """
# Galaxy Docker image

FROM bgruening/galaxy-stable

MAINTAINER Bjoern A. Gruning, bjoern.gruening@gmail.com

WORKDIR /galaxy-central

${selected_repositories}

# Mark folders as imported from the host.
VOLUME ["/export/", "/data/", "/var/lib/docker"]

# Expose port 80 (webserver), 21 (FTP server), 8800 (Proxy), 9001 (Galaxy report app)
EXPOSE :80
EXPOSE :21
EXPOSE :8800
EXPOSE :9001

# Autostart script that is invoked during container start
CMD ["/usr/bin/startup"]
"""

SELECTED_REPOSITORIES_TEMPLATE = """
RUN install-repository "--url ${tool_shed_url} -o ${repository_owner} --name ${repository_name}"
"""


def get_file_type_str(changeset_revision, file_type):
    if file_type == "zip":
        file_type_str = f"{changeset_revision}.zip"
    elif file_type == "bz2":
        file_type_str = f"{changeset_revision}.tar.bz2"
    elif file_type == "gz":
        file_type_str = f"{changeset_revision}.tar.gz"
    else:
        file_type_str = ""
    return file_type_str


def move_file(current_dir, source, destination, rename_to=None):
    source_path = os.path.abspath(os.path.join(current_dir, source))
    destination_directory = os.path.join(destination)
    if rename_to is not None:
        destination_path = os.path.join(destination_directory, rename_to)
    else:
        source_file = os.path.basename(source_path)
        destination_path = os.path.join(destination_directory, source_file)
    if not os.path.exists(destination_directory):
        os.makedirs(destination_directory)
    shutil.move(source_path, destination_path)


def remove_dir(dir):
    """Attempt to remove a directory from disk."""
    if dir:
        if os.path.exists(dir):
            try:
                shutil.rmtree(dir)
            except Exception:
                pass


def size_string(raw_text, size=MAX_DISPLAY_SIZE):
    """Return a subset of a string (up to MAX_DISPLAY_SIZE) translated to a safe string for display in a browser."""
    if raw_text and len(raw_text) >= size:
        large_str = (
            f"\nFile contents truncated because file size is larger than maximum viewing size of {nice_size(size)}\n"
        )
        raw_text = f"{raw_text[0:size]}{large_str}"
    return raw_text or ""


def stringify(list):
    if list:
        return ",".join(list)
    return ""


def strip_path(fpath):
    """Attempt to strip the path from a file name."""
    if not fpath:
        return fpath
    try:
        file_path, file_name = os.path.split(fpath)
    except Exception:
        file_name = fpath
    return file_name


def to_html_string(text):
    """Translates the characters in text to an html string"""
    if text:
        try:
            text = unicodify(text)
        except UnicodeDecodeError as e:
            return f"Error decoding string: {str(e)}"
        text = str(markupsafe.escape(text))
        text = text.replace("\n", "<br/>")
        text = text.replace("    ", "&nbsp;&nbsp;&nbsp;&nbsp;")
        text = text.replace(" ", "&nbsp;")
    return text


__all__ = (
    "CHUNK_SIZE",
    "DOCKER_IMAGE_TEMPLATE",
    "get_file_type_str",
    "INSTALLATION_LOG",
    "MAX_DISPLAY_SIZE",
    "MAXDIFFSIZE",
    "NO_OUTPUT_TIMEOUT",
    "move_file",
    "remove_dir",
    "SELECTED_REPOSITORIES_TEMPLATE",
    "size_string",
    "stringify",
    "strip_path",
    "to_html_string",
)



#!/usr/bin/env python
# By, Guruprasad Ananda.
from __future__ import print_function

import optparse
import re


def __main__():
    parser = optparse.OptionParser()
    parser.add_option("--strip", action="store_true", help="strip leading and trailing whitespaces")
    parser.add_option("--condense", action="store_true", help="condense consecutive delimiters")
    (options, args) = parser.parse_args()
    if len(args) != 3:
        parser.error("usage: convert_characters.py infile from_char outfile")

    char_dict = {"T": "\t", "s": r"\s", "Dt": r"\.", "C": ",", "D": "-", "U": "_", "P": r"\|", "Co": ":", "Sc": ";"}
    # regexp to match 1 or more occurences.
    from_char = args[1]
    from_ch = char_dict[from_char]
    if options.condense:
        from_ch += "+"

    skipped = 0
    with open(args[0]) as fin:
        with open(args[2], "w") as fout:
            for line in fin:
                if options.strip:
                    line = line.strip()
                else:
                    line = line.rstrip("\n")
                try:
                    fout.write("%s\n" % (re.sub(from_ch, "\t", line)))
                except Exception:
                    skipped += 1

    if skipped:
        print("Skipped %d lines as invalid." % skipped)


if __name__ == "__main__":
    __main__()



from sqlalchemy import (
    false,
    select,
    update,
)
from typing_extensions import Protocol

from galaxy.model.base import transaction
from galaxy.structured_app import BasicSharedApp


class IsUserModel(Protocol):
    id: int


class ApiKeyManager:
    def __init__(self, app: BasicSharedApp):
        self.app = app
        self.session = self.app.model.context

    def get_api_key(self, user: IsUserModel):
        APIKeys = self.app.model.APIKeys
        stmt = select(APIKeys).filter_by(user_id=user.id, deleted=False).order_by(APIKeys.create_time.desc()).limit(1)
        return self.session.scalars(stmt).first()

    def create_api_key(self, user: IsUserModel):
        guid = self.app.security.get_new_guid()
        new_key = self.app.model.APIKeys()
        new_key.user_id = user.id
        new_key.key = guid
        self.session.add(new_key)
        with transaction(self.session):
            self.session.commit()
        return new_key

    def get_or_create_api_key(self, user: IsUserModel) -> str:
        # Logic Galaxy has always used - but it would appear to have a race
        # condition. Worth fixing? Would kind of need a message queue to fix
        # in multiple process mode.
        api_key = self.get_api_key(user)
        key = api_key.key if api_key else self.create_api_key(user).key
        return key

    def delete_api_key(self, user: IsUserModel) -> None:
        """Marks the current user API key as deleted."""
        # Before it was possible to create multiple API keys for the same user although they were not considered valid
        # So all non-deleted keys are marked as deleted for backward compatibility
        self._mark_all_api_keys_as_deleted(user.id)
        with transaction(self.session):
            self.session.commit()

    def _mark_all_api_keys_as_deleted(self, user_id: int):
        APIKeys = self.app.model.APIKeys
        stmt = (
            update(APIKeys)
            .where(APIKeys.user_id == user_id)
            .where(APIKeys.deleted == false())
            .values(deleted=True)
            .execution_options(synchronize_session="evaluate")
        )
        return self.session.execute(stmt)



"""
Constructs for grouping tool parameters
"""

import io
import logging
import os
import unicodedata
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    TYPE_CHECKING,
)

from galaxy.datatypes import data
from galaxy.exceptions import (
    AdminRequiredException,
    ConfigDoesNotAllowException,
)
from galaxy.files.uris import stream_to_file
from galaxy.util import (
    asbool,
    inflector,
    relpath,
    sanitize_for_filename,
)
from galaxy.util.bunch import Bunch
from galaxy.util.dictifiable import Dictifiable
from galaxy.util.expressions import ExpressionContext

if TYPE_CHECKING:
    from galaxy.tools import Tool
    from galaxy.tools.parameter.basic import ToolParameter

log = logging.getLogger(__name__)
URI_PREFIXES = [
    f"{x}://"
    for x in [
        "http",
        "https",
        "ftp",
        "file",
        "gxfiles",
        "gximport",
        "gxuserimport",
        "gxftp",
        "drs",
        "invenio",
        "zenodo",
    ]
]


class Group(Dictifiable):
    dict_collection_visible_keys = ["name", "type"]
    type: str

    def __init__(self):
        self.name = None

    @property
    def visible(self):
        return True

    def value_to_basic(self, value, app, use_security=False):
        """
        Convert value to a (possibly nested) representation using only basic
        types (dict, list, tuple, string_types, int, long, float, bool, None)
        """
        return value

    def value_from_basic(self, value, app, ignore_errors=False):
        """
        Convert a basic representation as produced by `value_to_basic` back
        into the preferred value form.
        """
        return value

    def get_initial_value(self, trans, context):
        """
        Return the initial state/value for this group
        """
        raise TypeError("Not implemented")

    def to_dict(self, trans):
        group_dict = super().to_dict()
        return group_dict


class Repeat(Group):
    dict_collection_visible_keys = ["name", "type", "title", "help", "default", "min", "max"]
    type = "repeat"

    def __init__(self):
        Group.__init__(self)
        self._title = None
        self.inputs = None
        self.help = None
        self.default = 0
        self.min = None
        self.max = None

    @property
    def title(self):
        return self._title or self.name

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def title_plural(self):
        return inflector.pluralize(self.title)

    @property
    def label(self):
        return f"Repeat ({self.title})"

    def value_to_basic(self, value, app, use_security=False):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = []
        for d in value:
            rval_dict = {}
            # Propogate __index__
            if "__index__" in d:
                rval_dict["__index__"] = d["__index__"]
            for input in self.inputs.values():
                if input.name in d:
                    rval_dict[input.name] = input.value_to_basic(d[input.name], app, use_security)
            rval.append(rval_dict)
        return rval

    def value_from_basic(self, value, app, ignore_errors=False):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = []
        try:
            for i, d in enumerate(value):
                rval_dict = {}
                # If the special __index__ key is not set, create it (for backward
                # compatibility)
                rval_dict["__index__"] = d.get("__index__", i)
                # Restore child inputs
                for input in self.inputs.values():
                    if ignore_errors and input.name not in d:
                        # If we do not have a value, and are ignoring errors, we simply
                        # do nothing. There will be no value for the parameter in the
                        # conditional's values dictionary.
                        pass
                    else:
                        rval_dict[input.name] = input.value_from_basic(d[input.name], app, ignore_errors)
                rval.append(rval_dict)
        except Exception as e:
            if not ignore_errors:
                raise e
        return rval

    def get_initial_value(self, trans, context):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = []
        for i in range(self.default):
            rval_dict = {"__index__": i}
            for input in self.inputs.values():
                rval_dict[input.name] = input.get_initial_value(trans, context)
            rval.append(rval_dict)
        return rval

    def to_dict(self, trans):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        repeat_dict = super().to_dict(trans)

        def input_to_dict(input):
            return input.to_dict(trans)

        repeat_dict["inputs"] = list(map(input_to_dict, self.inputs.values()))
        return repeat_dict


class Section(Group):
    dict_collection_visible_keys = ["name", "type", "title", "help", "expanded"]
    type = "section"

    def __init__(self):
        Group.__init__(self)
        self.title = None
        self.inputs = None
        self.help = None
        self.expanded = False

    @property
    def title_plural(self):
        return inflector.pluralize(self.title)

    @property
    def label(self):
        return f"Section ({self.title})"

    def value_to_basic(self, value, app, use_security=False):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = {}
        for input in self.inputs.values():
            if input.name in value:  # parameter might be absent in unverified workflow
                rval[input.name] = input.value_to_basic(value[input.name], app, use_security)
        return rval

    def value_from_basic(self, value, app, ignore_errors=False):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = {}
        try:
            for input in self.inputs.values():
                if not ignore_errors or input.name in value:
                    rval[input.name] = input.value_from_basic(value[input.name], app, ignore_errors)
        except Exception as e:
            if not ignore_errors:
                raise e
        return rval

    def get_initial_value(self, trans, context):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval: Dict[str, Any] = {}
        child_context = ExpressionContext(rval, context)
        for child_input in self.inputs.values():
            rval[child_input.name] = child_input.get_initial_value(trans, child_context)
        return rval

    def to_dict(self, trans):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        section_dict = super().to_dict(trans)

        def input_to_dict(input):
            return input.to_dict(trans)

        section_dict["inputs"] = list(map(input_to_dict, self.inputs.values()))
        return section_dict


class Dataset(Bunch):
    type: str
    file_type: str
    dbkey: str
    datatype: data.Data
    warnings: List[str]
    metadata: Dict[str, str]
    composite_files: Dict[str, Optional[str]]
    uuid: Optional[str]
    tag_using_filenames: Optional[str]
    tags: Optional[str]
    name: str
    primary_file: str
    to_posix_lines: bool
    auto_decompress: bool
    ext: str
    space_to_tab: bool


class UploadDataset(Group):
    type = "upload_dataset"

    def __init__(self):
        Group.__init__(self)
        self.title = None
        self.inputs = None
        self.file_type_name = "file_type"
        self.default_file_type = "txt"
        self.file_type_to_ext = {"auto": self.default_file_type}
        self.metadata_ref = "files_metadata"

    def get_composite_dataset_name(self, context):
        # FIXME: HACK
        # Special case of using 'base_name' metadata for use as Dataset name needs to be done in a General Fashion, as defined within a particular Datatype.

        # We get two different types of contexts here, one straight from submitted parameters, the other after being parsed into tool inputs
        dataset_name = context.get("files_metadata|base_name", None)
        if dataset_name is None:
            dataset_name = context.get("files_metadata", {}).get("base_name", None)
        if dataset_name is None:
            filenames = []
            for composite_file in context.get("files", []):
                if not composite_file.get("ftp_files", ""):
                    filenames.append((composite_file.get("file_data") or {}).get("filename", ""))
                else:
                    filenames.append(composite_file.get("ftp_files", [])[0])
            dataset_name = os.path.commonprefix(filenames).rstrip(".") or None
        if dataset_name is None:
            dataset_name = f"Uploaded Composite Dataset ({self.get_file_type(context)})"
        return dataset_name

    def get_file_base_name(self, context):
        fd = context.get("files_metadata|base_name", "Galaxy_Composite_file")
        return fd

    def get_file_type(self, context, parent_context=None):
        file_type = context.get(self.file_type_name, None)
        if file_type == "":
            if parent_context:
                file_type = parent_context.get(self.file_type_name, self.default_file_type)
            else:
                file_type = self.default_file_type
        return file_type

    def get_dbkey(self, context, parent_context=None):
        dbkey = context.get("dbkey", None)
        if dbkey == "":
            if parent_context:
                dbkey = parent_context.get("dbkey", dbkey)
        return dbkey

    def get_datatype_ext(self, trans, context, parent_context=None):
        ext = self.get_file_type(context, parent_context=parent_context)
        if ext in self.file_type_to_ext:
            ext = self.file_type_to_ext[
                ext
            ]  # when using autodetect, we will use composite info from 'text', i.e. only the main file
        return ext

    def get_datatype(self, trans, context, parent_context=None):
        ext = self.get_datatype_ext(trans, context, parent_context=parent_context)
        return trans.app.datatypes_registry.get_datatype_by_extension(ext)

    @property
    def title_plural(self):
        return inflector.pluralize(self.title)

    def group_title(self, context):
        return f"{self.title} ({context.get(self.file_type_name, self.default_file_type)})"

    def title_by_index(self, trans, index, context):
        d_type = self.get_datatype(trans, context)
        for i, (composite_name, composite_file) in enumerate(d_type.writable_files.items()):
            if i == index:
                rval = composite_name
                if composite_file.description:
                    rval = f"{rval} ({composite_file.description})"
                if composite_file.optional:
                    rval = f"{rval} [optional]"
                return rval
        if index < self.get_file_count(trans, context):
            return "Extra primary file"
        return None

    def value_to_basic(self, value, app, use_security=False):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = []
        for d in value:
            rval_dict = {}
            # Propogate __index__
            if "__index__" in d:
                rval_dict["__index__"] = d["__index__"]
            for input in self.inputs.values():
                if input.name in d:
                    rval_dict[input.name] = input.value_to_basic(d[input.name], app, use_security)
            rval.append(rval_dict)
        return rval

    def value_from_basic(self, value, app, ignore_errors=False):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        rval = []
        for i, d in enumerate(value):
            try:
                rval_dict = {}
                # If the special __index__ key is not set, create it (for backward
                # compatibility)
                rval_dict["__index__"] = d.get("__index__", i)
                # Restore child inputs
                for input in self.inputs.values():
                    if ignore_errors and input.name not in d:  # this wasn't tested
                        rval_dict[input.name] = input.get_initial_value(None, d)
                    else:
                        rval_dict[input.name] = input.value_from_basic(d[input.name], app, ignore_errors)
                rval.append(rval_dict)
            except Exception as e:
                if not ignore_errors:
                    raise e
        return rval

    def get_file_count(self, trans, context):
        file_count = context.get("file_count", "auto")
        if file_count == "auto":
            d_type = self.get_datatype(trans, context)
            return len(d_type.writable_files) if d_type else 1
        else:
            return int(file_count)

    def get_initial_value(self, trans, context):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        file_count = self.get_file_count(trans, context)
        rval = []
        for i in range(file_count):
            rval_dict = {}
            rval_dict["__index__"] = i  # create __index__
            for input in self.inputs.values():
                rval_dict[input.name] = input.get_initial_value(trans, context)
            rval.append(rval_dict)
        return rval

    def get_uploaded_datasets(self, trans, context, override_name=None, override_info=None):
        def get_data_file_filename(data_file, override_name=None, override_info=None, purge=True):
            dataset_name = override_name

            def get_file_name(file_name):
                file_name = file_name.split("\\")[-1]
                file_name = file_name.split("/")[-1]
                return file_name

            try:
                # Use the existing file
                if not dataset_name and "filename" in data_file:
                    dataset_name = get_file_name(data_file["filename"])
                return Bunch(type="file", path=data_file["local_filename"], name=dataset_name, purge_source=purge)
            except Exception:
                # The uploaded file should've been persisted by the upload tool action
                return Bunch(type=None, path=None, name=None)

        def get_url_paste_urls_or_filename(group_incoming, override_name=None, override_info=None):
            if (url_paste_file := group_incoming.get("url_paste", None)) is not None:
                url_paste = open(url_paste_file).read()

                def start_of_url(content):
                    start_of_url_paste = content.lstrip()[0:10].lower()
                    looks_like_url = False
                    for url_prefix in URI_PREFIXES:
                        if start_of_url_paste.startswith(url_prefix):
                            looks_like_url = True
                            break

                    return looks_like_url

                if start_of_url(url_paste):
                    for line in url_paste.replace("\r", "").split("\n"):
                        line = line.strip()
                        if line:
                            if not start_of_url(line):
                                continue  # non-url line, ignore

                            if "file://" in line:
                                if not trans.user_is_admin:
                                    raise AdminRequiredException()
                                elif not trans.app.config.allow_path_paste:
                                    raise ConfigDoesNotAllowException()
                                upload_path = line[len("file://") :]
                                dataset_name = os.path.basename(upload_path)
                            else:
                                dataset_name = line

                            if override_name:
                                dataset_name = override_name
                            yield Bunch(type="url", path=line, name=dataset_name)
                else:
                    dataset_name = "Pasted Entry"  # we need to differentiate between various url pastes here
                    if override_name:
                        dataset_name = override_name
                    yield Bunch(type="file", path=url_paste_file, name=dataset_name)

        def get_one_filename(context):
            data_file = context["file_data"]
            url_paste = context["url_paste"]
            ftp_files = context["ftp_files"]
            name = context.get("NAME", None)
            info = context.get("INFO", None)
            uuid = context.get("uuid", None) or None  # Turn '' to None
            file_type = context.get("file_type", None)
            dbkey = self.get_dbkey(context)
            warnings = []
            to_posix_lines = False
            if context.get("to_posix_lines", None) not in ["None", None, False]:
                to_posix_lines = True
            auto_decompress = False
            if context.get("auto_decompress", None) not in ["None", None, False]:
                auto_decompress = True
            space_to_tab = False
            if context.get("space_to_tab", None) not in ["None", None, False]:
                space_to_tab = True
            file_bunch = get_data_file_filename(data_file, override_name=name, override_info=info)
            if file_bunch.path:
                if url_paste is not None and url_paste.strip():
                    warnings.append("All file contents specified in the paste box were ignored.")
                if ftp_files:
                    warnings.append("All FTP uploaded file selections were ignored.")
            elif url_paste is not None and url_paste.strip():  # we need to use url_paste
                for file_bunch in get_url_paste_urls_or_filename(context, override_name=name, override_info=info):
                    if file_bunch.path:
                        break
                if file_bunch.path and ftp_files is not None:
                    warnings.append("All FTP uploaded file selections were ignored.")
            elif ftp_files is not None and trans.user is not None:  # look for files uploaded via FTP
                user_ftp_dir = trans.user_ftp_dir
                assert not os.path.islink(user_ftp_dir), "User FTP directory cannot be a symbolic link"
                for dirpath, _dirnames, filenames in os.walk(user_ftp_dir):
                    for filename in filenames:
                        for ftp_filename in ftp_files:
                            if ftp_filename == filename:
                                path = relpath(os.path.join(dirpath, filename), user_ftp_dir)
                                if not os.path.islink(os.path.join(dirpath, filename)):
                                    ftp_data_file = {
                                        "local_filename": os.path.abspath(os.path.join(user_ftp_dir, path)),
                                        "filename": os.path.basename(path),
                                    }
                                    purge = getattr(trans.app.config, "ftp_upload_purge", True)
                                    file_bunch = get_data_file_filename(
                                        ftp_data_file,
                                        override_name=name,
                                        override_info=info,
                                        purge=purge,
                                    )
                                    if file_bunch.path:
                                        break
                        if file_bunch.path:
                            break
                    if file_bunch.path:
                        break
            file_bunch.to_posix_lines = to_posix_lines
            file_bunch.auto_decompress = auto_decompress
            file_bunch.space_to_tab = space_to_tab
            file_bunch.uuid = uuid
            if file_type is not None:
                file_bunch.file_type = file_type
            if dbkey is not None:
                file_bunch.dbkey = dbkey
            return file_bunch, warnings

        def get_filenames(context):
            rval = []
            data_file = context["file_data"]
            ftp_files = context["ftp_files"]
            uuid = context.get("uuid", None) or None  # Turn '' to None
            name = context.get("NAME", None)
            info = context.get("INFO", None)
            file_type = context.get("file_type", None)
            dbkey = self.get_dbkey(context)
            to_posix_lines = False
            if context.get("to_posix_lines", None) not in ["None", None, False]:
                to_posix_lines = True
            auto_decompress = False
            if context.get("auto_decompress", None) not in ["None", None, False]:
                auto_decompress = True
            space_to_tab = False
            if context.get("space_to_tab", None) not in ["None", None, False]:
                space_to_tab = True
            file_bunch = get_data_file_filename(data_file, override_name=name, override_info=info)
            file_bunch.uuid = uuid
            if file_bunch.path:
                file_bunch.to_posix_lines = to_posix_lines
                file_bunch.auto_decompress = auto_decompress
                file_bunch.space_to_tab = space_to_tab
                if file_type is not None:
                    file_bunch.file_type = file_type
                if dbkey is not None:
                    file_bunch.dbkey = dbkey

                rval.append(file_bunch)
            for file_bunch in get_url_paste_urls_or_filename(context, override_name=name, override_info=info):
                if file_bunch.path:
                    file_bunch.uuid = uuid
                    file_bunch.to_posix_lines = to_posix_lines
                    file_bunch.auto_decompress = auto_decompress
                    file_bunch.space_to_tab = space_to_tab
                    if file_type is not None:
                        file_bunch.file_type = file_type
                    if dbkey is not None:
                        file_bunch.dbkey = dbkey

                    rval.append(file_bunch)
            # look for files uploaded via FTP
            valid_files = []
            if ftp_files is not None:
                # Normalize input paths to ensure utf-8 encoding is normal form c.
                # This allows for comparison when the filesystem uses a different encoding than the browser.
                ftp_files = [unicodedata.normalize("NFC", f) for f in ftp_files if isinstance(f, str)]
                if trans.user is None:
                    log.warning(f"Anonymous user passed values in ftp_files: {ftp_files}")
                    ftp_files = []
                    # TODO: warning to the user (could happen if session has become invalid)
                else:
                    user_ftp_dir = trans.user_ftp_dir
                    assert not os.path.islink(user_ftp_dir), "User FTP directory cannot be a symbolic link"
                    for dirpath, _dirnames, filenames in os.walk(user_ftp_dir):
                        for filename in filenames:
                            path = relpath(os.path.join(dirpath, filename), user_ftp_dir)
                            if not os.path.islink(os.path.join(dirpath, filename)):
                                # Normalize filesystem paths
                                if isinstance(path, str):
                                    valid_files.append(unicodedata.normalize("NFC", path))
                                else:
                                    valid_files.append(path)

            else:
                ftp_files = []
            for ftp_file in ftp_files:
                if ftp_file not in valid_files:
                    log.warning(f"User passed an invalid file path in ftp_files: {ftp_file}")
                    continue
                    # TODO: warning to the user (could happen if file is already imported)
                ftp_data_file = {
                    "local_filename": os.path.abspath(os.path.join(user_ftp_dir, ftp_file)),
                    "filename": os.path.basename(ftp_file),
                }
                purge = getattr(trans.app.config, "ftp_upload_purge", True)
                file_bunch = get_data_file_filename(ftp_data_file, override_name=name, override_info=info, purge=purge)
                if file_bunch.path:
                    file_bunch.to_posix_lines = to_posix_lines
                    file_bunch.auto_decompress = auto_decompress
                    file_bunch.space_to_tab = space_to_tab
                    if file_type is not None:
                        file_bunch.file_type = file_type
                    if dbkey is not None:
                        file_bunch.dbkey = dbkey
                    rval.append(file_bunch)
            return rval

        file_type = self.get_file_type(context)
        file_count = self.get_file_count(trans, context)
        d_type = self.get_datatype(trans, context)
        dbkey = self.get_dbkey(context)
        tag_using_filenames = context.get("tag_using_filenames", False)
        tags = context.get("tags", False)
        force_composite = asbool(context.get("force_composite", "False"))
        writable_files = d_type.writable_files
        writable_files_offset = 0
        groups_incoming = [None for _ in range(file_count)]
        for i, group_incoming in enumerate(context.get(self.name, [])):
            i = int(group_incoming.get("__index__", i))
            groups_incoming[i] = group_incoming
        if d_type.composite_type is not None or force_composite:
            # handle uploading of composite datatypes
            # Only one Dataset can be created
            dataset = Dataset()
            dataset.type = "composite"
            dataset.file_type = file_type
            dataset.dbkey = dbkey
            dataset.datatype = d_type
            dataset.warnings = []
            dataset.metadata = {}
            dataset.composite_files = {}
            dataset.uuid = None
            dataset.tag_using_filenames = None
            dataset.tags = None
            # load metadata
            files_metadata = context.get(self.metadata_ref, {})
            metadata_name_substition_default_dict = {
                composite_file.substitute_name_with_metadata: d_type.metadata_spec[
                    composite_file.substitute_name_with_metadata
                ].default
                for composite_file in d_type.composite_files.values()
                if composite_file.substitute_name_with_metadata
            }
            for meta_name, meta_spec in d_type.metadata_spec.items():
                if meta_spec.set_in_upload:
                    if meta_name in files_metadata:
                        meta_value = files_metadata[meta_name]
                        if meta_name in metadata_name_substition_default_dict:
                            meta_value = sanitize_for_filename(
                                meta_value, default=metadata_name_substition_default_dict[meta_name]
                            )
                        dataset.metadata[meta_name] = meta_value
            dataset.name = self.get_composite_dataset_name(context)
            if dataset.datatype.composite_type == "auto_primary_file":
                # replace sniff here with just creating an empty file
                temp_name = stream_to_file(
                    io.StringIO(d_type.generate_primary_file(dataset)), prefix="upload_auto_primary_file"
                )
                dataset.primary_file = temp_name
                dataset.to_posix_lines = True
                dataset.auto_decompress = True
                dataset.space_to_tab = False
            else:
                file_bunch, warnings = get_one_filename(groups_incoming[0])
                writable_files_offset = 1
                dataset.primary_file = file_bunch.path
                dataset.to_posix_lines = file_bunch.to_posix_lines
                dataset.auto_decompress = file_bunch.auto_decompress
                dataset.space_to_tab = file_bunch.space_to_tab
                if file_bunch.file_type:
                    dataset.file_type = file_type
                if file_bunch.dbkey:
                    dataset.dbkey = dbkey
                dataset.warnings.extend(warnings)
            if dataset.primary_file is None:  # remove this before finish, this should create an empty dataset
                raise Exception("No primary dataset file was available for composite upload")
            if not force_composite:
                keys = [value.name for value in writable_files.values()]
            else:
                keys = [str(index) for index in range(file_count)]
            for i, group_incoming in enumerate(groups_incoming[writable_files_offset:]):
                key = keys[i + writable_files_offset]
                if (
                    not force_composite
                    and group_incoming is None
                    and not writable_files[list(writable_files.keys())[keys.index(key)]].optional
                ):
                    dataset.warnings.append(f"A required composite file ({key}) was not specified.")
                    dataset.composite_files[key] = None
                else:
                    file_bunch, warnings = get_one_filename(group_incoming)
                    dataset.warnings.extend(warnings)
                    if file_bunch.path:
                        if force_composite:
                            key = group_incoming.get("NAME") or i
                        dataset.composite_files[key] = file_bunch.__dict__
                    elif not force_composite:
                        dataset.composite_files[key] = None
                        if not writable_files[list(writable_files.keys())[keys.index(key)]].optional:
                            dataset.warnings.append(f"A required composite file ({key}) was not specified.")
            return [dataset]
        else:
            rval = []
            for i, file_contexts in enumerate(context[self.name]):
                datasets = get_filenames(file_contexts)
                for dataset in datasets:
                    override_file_type = self.get_file_type(context[self.name][i], parent_context=context)
                    d_type = self.get_datatype(trans, context[self.name][i], parent_context=context)
                    dataset.file_type = override_file_type
                    dataset.datatype = d_type
                    dataset.ext = self.get_datatype_ext(trans, context[self.name][i], parent_context=context)
                    dataset.dbkey = self.get_dbkey(context[self.name][i], parent_context=context)
                    dataset.tag_using_filenames = tag_using_filenames
                    dataset.tags = tags
                    rval.append(dataset)
            return rval


class Conditional(Group):
    type = "conditional"
    value_from: Callable[[ExpressionContext, "Conditional", "Tool"], Mapping[str, str]]

    def __init__(self):
        Group.__init__(self)
        self.test_param: Optional[ToolParameter] = None
        self.cases = []
        self.value_ref = None
        self.value_ref_in_group = True  # When our test_param is not part of the conditional Group, this is False

    @property
    def label(self):
        return f"Conditional ({self.name})"

    def get_current_case(self, value):
        if self.test_param is None:
            raise Exception("Must set 'test_param' attribute to use.")
        # Convert value to user representation
        str_value = self.test_param.to_param_dict_string(value)
        # Find the matching case
        for index, case in enumerate(self.cases):
            if str_value == case.value:
                return index
        raise ValueError("No case matched value:", self.name, str_value)

    def value_to_basic(self, value, app, use_security=False):
        if self.test_param is None:
            raise Exception("Must set 'test_param' attribute to use.")
        rval = {}
        rval[self.test_param.name] = self.test_param.value_to_basic(value[self.test_param.name], app)
        current_case = rval["__current_case__"] = self.get_current_case(value[self.test_param.name])
        for input in self.cases[current_case].inputs.values():
            if input.name in value:  # parameter might be absent in unverified workflow
                rval[input.name] = input.value_to_basic(value[input.name], app, use_security=use_security)
        return rval

    def value_from_basic(self, value, app, ignore_errors=False):
        if self.test_param is None:
            raise Exception("Must set 'test_param' attribute to use.")
        rval = {}
        try:
            rval[self.test_param.name] = self.test_param.value_from_basic(
                value.get(self.test_param.name), app, ignore_errors
            )
            current_case = rval["__current_case__"] = self.get_current_case(rval[self.test_param.name])
            # Inputs associated with current case
            for input in self.cases[current_case].inputs.values():
                # If we do not have a value, and are ignoring errors, we simply
                # do nothing. There will be no value for the parameter in the
                # conditional's values dictionary.
                if not ignore_errors or input.name in value:
                    rval[input.name] = input.value_from_basic(value[input.name], app, ignore_errors)
        except Exception as e:
            if not ignore_errors:
                raise e
        return rval

    def get_initial_value(self, trans, context):
        if self.test_param is None:
            raise Exception("Must set 'test_param' attribute to use.")
        # State for a conditional is a plain dictionary.
        rval = {}
        # Get the default value for the 'test element' and use it
        # to determine the current case
        test_value = self.test_param.get_initial_value(trans, context)
        current_case = self.get_current_case(test_value)
        # Store the current case in a special value
        rval["__current_case__"] = current_case
        # Store the value of the test element
        rval[self.test_param.name] = test_value
        # Fill in state for selected case
        child_context = ExpressionContext(rval, context)
        for child_input in self.cases[current_case].inputs.values():
            rval[child_input.name] = child_input.get_initial_value(trans, child_context)
        return rval

    def to_dict(self, trans):
        if self.test_param is None:
            raise Exception("Must set 'test_param' attribute to use.")
        cond_dict = super().to_dict(trans)

        def nested_to_dict(input):
            return input.to_dict(trans)

        cond_dict["cases"] = list(map(nested_to_dict, self.cases))
        cond_dict["test_param"] = nested_to_dict(self.test_param)
        return cond_dict


class ConditionalWhen(Dictifiable):
    dict_collection_visible_keys = ["value"]

    def __init__(self):
        self.value = None
        self.inputs = None

    def to_dict(self, trans):
        if self.inputs is None:
            raise Exception("Must set 'inputs' attribute to use.")
        when_dict = super().to_dict()

        def input_to_dict(input):
            return input.to_dict(trans)

        when_dict["inputs"] = list(map(input_to_dict, self.inputs.values()))
        return when_dict



import json
import os
import shutil
import tempfile
import traceback
from typing import (
    Callable,
    NamedTuple,
)

from galaxy.datatypes.registry import Registry
from galaxy.files import ConfiguredFileSources
from galaxy.job_execution.compute_environment import SharedComputeEnvironment
from galaxy.job_execution.setup import JobIO
from galaxy.managers.dbkeys import GenomeBuilds
from galaxy.metadata.set_metadata import (
    get_metadata_params,
    get_object_store,
    validate_and_load_datatypes_config,
)
from galaxy.model import store
from galaxy.model.store import SessionlessContext
from galaxy.objectstore import BaseObjectStore
from galaxy.structured_app import MinimalToolApp
from galaxy.tools import (
    create_tool_from_representation,
    evaluation,
)
from galaxy.tools.data import (
    from_dict,
    ToolDataTableManager,
)
from galaxy.util.bunch import Bunch


class ToolAppConfig(NamedTuple):
    name: str
    tool_data_path: str
    galaxy_data_manager_data_path: str
    nginx_upload_path: str
    len_file_path: str
    builds_file_path: str
    root: str
    is_admin_user: Callable
    admin_users: list = []


class ToolApp(MinimalToolApp):
    """Dummy App that allows loading tools"""

    name = "tool_app"
    is_webapp = False

    def __init__(
        self,
        sa_session: SessionlessContext,
        tool_app_config: ToolAppConfig,
        datatypes_registry: Registry,
        object_store: BaseObjectStore,
        tool_data_table_manager: ToolDataTableManager,
        file_sources: ConfiguredFileSources,
    ):
        # For backward compatibility we need both context and session attributes that point to sa_session.
        self.model = Bunch(context=sa_session, session=sa_session)
        self.config = tool_app_config
        self.datatypes_registry = datatypes_registry
        self.object_store = object_store
        self.genome_builds = GenomeBuilds(self)
        self.tool_data_tables = tool_data_table_manager
        self.file_sources = file_sources
        self.biotools_metadata_source = None
        self.security = None  # type: ignore[assignment]


def main(TMPDIR, WORKING_DIRECTORY, IMPORT_STORE_DIRECTORY):
    metadata_params = get_metadata_params(WORKING_DIRECTORY)
    datatypes_config = metadata_params["datatypes_config"]
    if not os.path.exists(datatypes_config):
        datatypes_config = os.path.join(WORKING_DIRECTORY, "configs", datatypes_config)
    datatypes_registry = validate_and_load_datatypes_config(datatypes_config)
    object_store = get_object_store(WORKING_DIRECTORY)
    import_store = store.imported_store_for_metadata(IMPORT_STORE_DIRECTORY)
    # TODO: clean up random places from which we read files in the working directory
    job_io = JobIO.from_json(os.path.join(IMPORT_STORE_DIRECTORY, "job_io.json"), sa_session=import_store.sa_session)
    tool_app_config = ToolAppConfig(
        name="tool_app",
        tool_data_path=job_io.tool_data_path,
        galaxy_data_manager_data_path=job_io.galaxy_data_manager_data_path,
        nginx_upload_path=TMPDIR,
        len_file_path=job_io.len_file_path,
        builds_file_path=job_io.builds_file_path,
        root=TMPDIR,
        is_admin_user=lambda _: job_io.user_context.is_admin,
    )
    with open(os.path.join(IMPORT_STORE_DIRECTORY, "tool_data_tables.json")) as data_tables_json:
        tdtm = from_dict(json.load(data_tables_json))
    app = ToolApp(
        sa_session=import_store.sa_session,
        tool_app_config=tool_app_config,
        datatypes_registry=datatypes_registry,
        object_store=object_store,
        tool_data_table_manager=tdtm,
        file_sources=job_io.file_sources,
    )
    # TODO: could try to serialize just a minimal tool variant instead of the whole thing ?
    tool = create_tool_from_representation(
        app=app,
        raw_tool_source=job_io.tool_source,
        tool_dir=job_io.tool_dir,
        tool_source_class=job_io.tool_source_class,
    )
    tool_evaluator = evaluation.RemoteToolEvaluator(
        app=app, tool=tool, job=job_io.job, local_working_directory=WORKING_DIRECTORY
    )
    tool_evaluator.set_compute_environment(compute_environment=SharedComputeEnvironment(job_io=job_io, job=job_io.job))
    with open(os.path.join(WORKING_DIRECTORY, "tool_script.sh"), "a") as out:
        command_line, version_command_line, extra_filenames, environment_variables, *_ = tool_evaluator.build()
        out.write(f'{version_command_line or ""}{command_line}')


if __name__ == "__main__":
    TMPDIR = tempfile.mkdtemp()
    WORKING_DIRECTORY = os.getcwd()
    WORKING_PARENT = os.path.join(WORKING_DIRECTORY, os.path.pardir)
    if not os.path.isdir("working") and os.path.isdir(os.path.join(WORKING_PARENT, "working")):
        # We're probably in pulsar
        WORKING_DIRECTORY = WORKING_PARENT
    METADATA_DIRECTORY = os.path.join(WORKING_DIRECTORY, "metadata")
    IMPORT_STORE_DIRECTORY = os.path.join(METADATA_DIRECTORY, "outputs_new")
    EXPORT_STORE_DIRECTORY = os.path.join(METADATA_DIRECTORY, "outputs_populated")
    try:
        main(TMPDIR, WORKING_DIRECTORY, IMPORT_STORE_DIRECTORY)
    except Exception:
        os.makedirs(EXPORT_STORE_DIRECTORY, exist_ok=True)
        with open(os.path.join(EXPORT_STORE_DIRECTORY, "traceback.txt"), "w") as out:
            out.write(traceback.format_exc())
        raise
    finally:
        shutil.rmtree(TMPDIR, ignore_errors=True)



"""The module describes the ``json`` error plugin."""

import json
import logging
import os
import tempfile

from galaxy.util import string_as_bool
from . import ErrorPlugin

log = logging.getLogger(__name__)


class JsonPlugin(ErrorPlugin):
    """Write error report to a JSON file."""

    plugin_type = "json"

    def __init__(self, **kwargs):
        self.app = kwargs["app"]
        self.verbose = string_as_bool(kwargs.get("verbose", False))
        self.redact_user_details_in_bugreport = self.app.config.redact_user_details_in_bugreport
        self.user_submission = string_as_bool(kwargs.get("user_submission", False))
        self.report_directory = kwargs.get("directory", tempfile.gettempdir())
        if not os.path.exists(self.report_directory):
            os.makedirs(self.report_directory)

    def submit_report(self, dataset, job, tool, **kwargs):
        """Write the report to a json file."""
        path = os.path.join(self.report_directory, str(dataset.id))
        with open(path, "w") as handle:
            data = {
                "info": job.info,
                "id": job.id,
                "command_line": job.command_line,
                "destination_id": job.destination_id,
                "stderr": job.stderr,
                "traceback": job.traceback,
                "exit_code": job.exit_code,
                "stdout": job.stdout,
                "handler": job.handler,
                "tool_version": job.tool_version,
                "tool_xml": str(tool.config_file) if tool else None,
            }
            if self.redact_user_details_in_bugreport:
                data["user"] = {"id": job.get_user().id}
            else:
                data["user"] = job.get_user().to_dict()
                if "email" in kwargs:
                    data["email"] = kwargs["email"]

            if "message" in kwargs:
                data["message"] = kwargs["message"]

            json.dump(data, handle, indent=2)
        return (f"Wrote error report to {path}", "success")


__all__ = ("JsonPlugin",)



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/ga4gh-discovery/ga4gh-service-info/v1.0.0/service-info.yaml#/paths/~1service-info
#   timestamp: 2022-12-20T21:01:57+00:00

from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import (
    AnyUrl,
    BaseModel,
    Field,
)


class Organization(BaseModel):
    name: str = Field(
        ..., description="Name of the organization responsible for the service", examples=["My organization"]
    )
    url: AnyUrl = Field(
        ..., description="URL of the website of the organization (RFC 3986 format)", examples=["https://example.com"]
    )


class ServiceType(BaseModel):
    group: str = Field(
        ...,
        description="Namespace in reverse domain name format. Use `org.ga4gh` for implementations compliant with official GA4GH specifications. For services with custom APIs not standardized by GA4GH, or implementations diverging from official GA4GH specifications, use a different namespace (e.g. your organization's reverse domain name).",
        examples=["org.ga4gh"],
    )
    artifact: str = Field(
        ...,
        description="Name of the API or GA4GH specification implemented. Official GA4GH types should be assigned as part of standards approval process. Custom artifacts are supported.",
        examples=["beacon"],
    )
    version: str = Field(
        ...,
        description="Version of the API or specification. GA4GH specifications use semantic versioning.",
        examples=["1.0.0"],
    )


class Service(BaseModel):
    id: str = Field(
        ...,
        description="Unique ID of this service. Reverse domain name notation is recommended, though not required. The identifier should attempt to be globally unique so it can be used in downstream aggregator services e.g. Service Registry.",
        examples=["org.ga4gh.myservice"],
    )
    name: str = Field(..., description="Name of this service. Should be human readable.", examples=["My project"])
    type: ServiceType
    description: Optional[str] = Field(
        None,
        description="Description of the service. Should be human readable and provide information about the service.",
        examples=["This service provides..."],
    )
    organization: Organization = Field(..., description="Organization providing the service")
    contactUrl: Optional[AnyUrl] = Field(
        None,
        description="URL of the contact for the provider of this service, e.g. a link to a contact form (RFC 3986 format), or an email (RFC 2368 format).",
        examples=["mailto:support@example.com"],
    )
    documentationUrl: Optional[AnyUrl] = Field(
        None,
        description="URL of the documentation of this service (RFC 3986 format). This should help someone learn how to use your service, including any specifics required to access data, e.g. authentication.",
        examples=["https://docs.myservice.example.com"],
    )
    createdAt: Optional[datetime] = Field(
        None,
        description="Timestamp describing when the service was first deployed and available (RFC 3339 format)",
        examples=["2019-06-04T12:58:19Z"],
    )
    updatedAt: Optional[datetime] = Field(
        None,
        description="Timestamp describing when the service was last updated (RFC 3339 format)",
        examples=["2019-06-04T12:58:19Z"],
    )
    environment: Optional[str] = Field(
        None,
        description="Environment the service is running in. Use this to distinguish between production, development and testing/staging deployments. Suggested values are prod, test, dev, staging. However this is advised and not enforced.",
        examples=["test"],
    )
    version: str = Field(
        ...,
        description="Version of the service being described. Semantic versioning is recommended, but other identifiers, such as dates or commit hashes, are also allowed. The version should be changed whenever the service is updated.",
        examples=["1.0.0"],
    )



"""
This tool implements the logic of sending data
from Galaxy to a cloud-based storage.

This tool depends on the CloudManager for configuring
a connection to a cloud-based resource provider. Also,
it leverages Cloudbridge (github.com/CloudVE/cloudbridge)
to send a dataset to a cloud-based storage.

"""

import argparse
import datetime
import json
import os
import sys
import time

from galaxy.exceptions import ObjectNotFound
from galaxy.managers.cloud import CloudManager

try:
    from cloudbridge.factory import (
        CloudProviderFactory,
        ProviderList,
    )
except ImportError:
    CloudProviderFactory = None
    ProviderList = None

NO_CLOUDBRIDGE_ERROR_MESSAGE = (
    "Cloud ObjectStore is configured, but no CloudBridge dependency available."
    "Please install CloudBridge or modify ObjectStore configuration."
)


def load_credential(credentials_file):
    print("[1/5 {}] Reading cloud authorization.".format(datetime.datetime.now().replace(microsecond=0)))
    with open(credentials_file) as f:
        credentials = f.read()
    os.remove(credentials_file)
    return json.loads(credentials)


def send(provider, credentials, bucket, object_label, filename, overwrite_existing):
    start_time = time.time()
    if not os.path.exists(filename):
        raise Exception("The file `{}` does not exist.".format(filename))
    if CloudProviderFactory is None:
        raise Exception(NO_CLOUDBRIDGE_ERROR_MESSAGE)

    print("[2/5 {}] Establishing a connection to {}.".format(datetime.datetime.now().replace(microsecond=0), provider))
    try:
        connection = CloudManager.configure_provider(provider, credentials)
    except Exception:
        print("Failed to establish the connection.")

    print("[3/5 {}] Accessing bucket {}.".format(datetime.datetime.now().replace(microsecond=0), bucket))
    bucket_obj = connection.storage.buckets.get(bucket)
    if bucket_obj is None:
        raise ObjectNotFound("Could not find the specified bucket `{}`.".format(bucket))
    if overwrite_existing is False and bucket_obj.objects.get(object_label) is not None:
        object_label += "_" + datetime.datetime.now().strftime("%y%m%d_%H%M%S")

    print("[4/5 {}] Creating object {}.".format(datetime.datetime.now().replace(microsecond=0), object_label))
    try:
        created_obj = bucket_obj.objects.create(object_label)
    except Exception:
        print("Failed to create the object.")

    print("[5/5 {}] Sending dataset.".format(datetime.datetime.now().replace(microsecond=0)))
    transfer_start_time = time.time()
    try:
        created_obj.upload_from_file(filename)
    except Exception:
        print("Failed to send the dataset.")

    print("Finished successfully.")
    print("Job runtime:\t{}".format(time.time() - start_time))
    print(
        "Transfer ET:\t{}\tSpeed:\t{}MB/sec".format(
            time.time() - transfer_start_time,
            round((os.path.getsize(filename) >> 20) / (time.time() - transfer_start_time), 3),
        )
    )


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--provider", type=str, required=True, help="Provider")

    parser.add_argument(
        "-b",
        "--bucket",
        type=str,
        required=True,
        help="The cloud-based storage bucket in which data should be written.",
    )

    parser.add_argument(
        "-o",
        "--object_label",
        type=str,
        required=True,
        help="The label of the object created on the cloud-based storage for the data to be persisted.",
    )

    parser.add_argument(
        "-f",
        "--filename",
        type=str,
        required=True,
        help="The (absolute) filename of the data to be persisted on the cloud-based storage.",
    )

    parser.add_argument(
        "-w",
        "--overwrite_existing",
        type=str,
        required=True,
        help="Sets if an object with the given `object_label` exists, this tool "
        "should overwrite it (true) or append a time stamp to avoid "
        "overwriting (false).",
    )

    parser.add_argument("--credentials_file", type=str, required=True, help="Credentials file")

    return parser.parse_args(args)


def __main__():
    args = parse_args(sys.argv[1:])
    overwrite_existing = args.overwrite_existing.lower() == "true"
    credentials = load_credential(args.credentials_file)
    send(args.provider, credentials, args.bucket, args.object_label, args.filename, overwrite_existing)


if __name__ == "__main__":
    sys.exit(__main__())



#!/usr/bin/env python
"""
Reads a list of intervals and a maf. Produces a new maf containing the
blocks or parts of blocks in the original that overlapped the intervals.

If a MAF file, not UID, is provided the MAF file is indexed before being processed.

NOTE: If two intervals overlap the same block it will be written twice.

usage: %prog maf_file [options]
   -d, --dbkey=d: Database key, ie hg17
   -c, --chromCol=c: Column of Chr
   -s, --startCol=s: Column of Start
   -e, --endCol=e: Column of End
   -S, --strandCol=S: Column of Strand
   -t, --mafType=t: Type of MAF source to use
   -m, --mafFile=m: Path of source MAF file, if not using cached version
   -I, --mafIndex=I: Path of precomputed source MAF file index, if not using cached version
   -i, --interval_file=i:       Input interval file
   -o, --output_file=o:      Output MAF file
   -p, --species=p: Species to include in output
   -P, --split_blocks_by_species=P: Split blocks by species
   -r, --remove_all_gap_columns=r: Remove all Gap columns
   -l, --indexLocation=l: Override default maf_index.loc file
   -z, --mafIndexFile=z: Directory of local maf index file ( maf_index.loc or maf_pairwise.loc )
"""
# Dan Blankenberg
from __future__ import print_function

import bx.align.maf
import bx.intervals.io
from bx.cookbook import doc_optparse

from galaxy.tools.util import maf_utilities


def __main__():
    index = index_filename = None

    # Parse Command Line
    options, args = doc_optparse.parse(__doc__)

    if options.dbkey:
        dbkey = options.dbkey
    else:
        dbkey = None
    if dbkey in [None, "?"]:
        maf_utilities.tool_fail(
            "You must specify a proper build in order to extract alignments. You can specify your genome build by clicking on the pencil icon associated with your interval file."
        )

    species = maf_utilities.parse_species_option(options.species)

    if options.chromCol:
        chromCol = int(options.chromCol) - 1
    else:
        maf_utilities.tool_fail(
            "Chromosome column not set, click the pencil icon in the history item to set the metadata attributes."
        )

    if options.startCol:
        startCol = int(options.startCol) - 1
    else:
        maf_utilities.tool_fail(
            "Start column not set, click the pencil icon in the history item to set the metadata attributes."
        )

    if options.endCol:
        endCol = int(options.endCol) - 1
    else:
        maf_utilities.tool_fail(
            "End column not set, click the pencil icon in the history item to set the metadata attributes."
        )

    if options.strandCol:
        strandCol = int(options.strandCol) - 1
    else:
        strandCol = -1

    if options.interval_file:
        interval_file = options.interval_file
    else:
        maf_utilities.tool_fail("Input interval file has not been specified.")

    if options.output_file:
        output_file = options.output_file
    else:
        maf_utilities.tool_fail("Output file has not been specified.")

    split_blocks_by_species = remove_all_gap_columns = False
    if options.split_blocks_by_species and options.split_blocks_by_species == "split_blocks_by_species":
        split_blocks_by_species = True
        if options.remove_all_gap_columns and options.remove_all_gap_columns == "remove_all_gap_columns":
            remove_all_gap_columns = True
    else:
        remove_all_gap_columns = True
    # Finish parsing command line

    # Open indexed access to MAFs
    if options.mafType:
        if options.indexLocation:
            index = maf_utilities.maf_index_by_uid(options.mafType, options.indexLocation)
        else:
            index = maf_utilities.maf_index_by_uid(options.mafType, options.mafIndexFile)
        if index is None:
            maf_utilities.tool_fail("The MAF source specified (%s) appears to be invalid." % (options.mafType))
    elif options.mafFile:
        index, index_filename = maf_utilities.open_or_build_maf_index(
            options.mafFile, options.mafIndex, species=[dbkey]
        )
        if index is None:
            maf_utilities.tool_fail("Your MAF file appears to be malformed.")
    else:
        maf_utilities.tool_fail("Desired source MAF type has not been specified.")

    # Create MAF writter
    out = bx.align.maf.Writer(open(output_file, "w"))

    # Iterate over input regions
    num_blocks = 0
    num_regions = None
    for num_regions, region in enumerate(  # noqa: B007
        bx.intervals.io.NiceReaderWrapper(
            open(interval_file),
            chrom_col=chromCol,
            start_col=startCol,
            end_col=endCol,
            strand_col=strandCol,
            fix_strand=True,
            return_header=False,
            return_comments=False,
        )
    ):
        src = maf_utilities.src_merge(dbkey, region.chrom)
        for block in index.get_as_iterator(src, region.start, region.end):
            if split_blocks_by_species:
                blocks = [
                    new_block
                    for new_block in maf_utilities.iter_blocks_split_by_species(block)
                    if maf_utilities.component_overlaps_region(new_block.get_component_by_src_start(dbkey), region)
                ]
            else:
                blocks = [block]
            for block in blocks:
                block = maf_utilities.chop_block_by_region(block, src, region)
                if block is not None:
                    if species is not None:
                        block = block.limit_to_species(species)
                    block = maf_utilities.orient_block_by_region(block, src, region)
                    if remove_all_gap_columns:
                        block.remove_all_gap_columns()
                    out.write(block)
                    num_blocks += 1

    # Close output MAF
    out.close()

    # remove index file if created during run
    maf_utilities.remove_temp_index_file(index_filename)

    if num_blocks:
        print("%i MAF blocks extracted for %i regions." % (num_blocks, (num_regions + 1)))
    elif num_regions is not None:
        print("No MAF blocks could be extracted for %i regions." % (num_regions + 1))
    else:
        print("No valid regions have been provided.")


if __name__ == "__main__":
    __main__()



import json
from typing import (
    Any,
    Dict,
)


class Node:
    """Node class of PhyloTree, which represents a CLAUDE in a phylogenetic tree"""

    def __init__(self, nodeName, **kwargs):
        """Creates a node and adds in the typical annotations"""
        self.name, self.id = nodeName, kwargs.get("id", 0)
        self.depth = kwargs.get("depth", 0)
        self.children = []

        self.isInternal = kwargs.get("isInternal", 0)
        self.length, self.bootstrap = kwargs.get("length", 0), kwargs.get("bootstrap", None)
        self.events = kwargs.get("events", "")

        self.parent = None

        # clean up boot strap values
        if self.bootstrap == -1:
            self.bootstrap = None

    def addChildNode(self, child):
        """Adds a child node to the current node"""
        if isinstance(child, Node):
            self.children.append(child)
        else:
            self.children += child

    def __str__(self):
        return f"{self.name} id:{str(self.id)}, depth: {str(self.depth)}"

    def toJson(self) -> Dict[str, Any]:
        """Converts the data in the node to a dict representation of json"""
        thisJson = {"name": self.name, "id": self.id, "depth": self.depth, "dist": self.length}
        thisJson = self.addChildrenToJson(thisJson)
        thisJson = self.addMiscToJson(thisJson)
        return thisJson

    def addChildrenToJson(self, jsonDict):
        """Needs a special method to addChildren, such that the key does not appear in the Jsondict when the children is empty
        this requirement is due to the layout algorithm used by d3 layout for hiding subtree"""
        if len(self.children) > 0:
            children = [node.toJson() for node in self.children]
            jsonDict["children"] = children
        return jsonDict

    def addMiscToJson(self, jsonDict):
        """Adds other misc attributes to json if they are present"""
        if not self.events == "":
            jsonDict["events"] = self.events
        if self.bootstrap is not None:
            jsonDict["bootstrap"] = self.bootstrap
        return jsonDict


class PhyloTree:
    """Standardized python based class to represent the phylogenetic tree parsed from different
    phylogenetic file formats."""

    def __init__(self):
        self.root, self.rootAttr = None, {}
        self.nodes = {}
        self.title = None
        self.id = 1

    def addAttributesToRoot(self, attrDict):
        """Adds attributes to root, but first we put it in a temp store and bind it with root when .toJson is called"""
        for key, value in attrDict.items():
            self.rootAttr[key] = value

    def makeNode(self, nodeName, **kwargs):
        """Called to make a node within PhyloTree, arbitrary kwargs can be passed to annotate nodes
        Tracks the number of nodes via internally incremented id"""
        kwargs["id"] = self.id
        self.id += 1
        return Node(nodeName, **kwargs)

    def addRoot(self, root: Node):
        """Creates a root for phyloTree"""
        assert isinstance(root, Node)
        root.parent = None
        self.root = root

    def generateJsonableDict(self):
        """Changes itself into a dictonary by recurssively calling the tojson on all its nodes. Think of it
        as a dict in an array of dict in an array of dict and so on..."""
        jsonTree: Dict[str, Any]
        if self.root:
            assert isinstance(self.root, Node)
            jsonTree = self.root.toJson()
            for key, value in self.rootAttr.items():
                # transfer temporary stored attr to root
                jsonTree[key] = value
        else:
            raise Exception("Root is not assigned!")
        return jsonTree


class Base_Parser:
    """Base parsers contain all the methods to handle phylogeny tree creation and
    converting the data to json that all parsers should have"""

    def __init__(self):
        self.phyloTrees = []

    def parseFile(self, filePath):
        """Base method that all phylogeny file parser should have"""
        raise Exception("Base method for phylogeny file parsers is not implemented")

    def toJson(self, jsonDict):
        """Convenience method to get a json string from a python json dict"""
        return json.dumps(jsonDict)

    def _writeJsonToFile(self, filepath, json):
        """Writes the file out to the system"""
        f = open(filepath, "w")
        f.writelines(json)
        f.close()



import logging

log = logging.getLogger(__name__)


class HistoryQuery:
    """An object for describing the collections to pull out of a history,
    used by DataCollectionToolParameter.
    """

    def __init__(self, **kwargs):
        self.collection_type_descriptions = kwargs.get("collection_type_descriptions", None)

    @staticmethod
    def from_collection_type(collection_type, collection_type_descriptions):
        kwargs = dict(collection_type_descriptions=[collection_type_descriptions.for_collection_type(collection_type)])
        return HistoryQuery(**kwargs)

    @staticmethod
    def from_collection_types(collection_types, collection_type_descriptions):
        if collection_types:
            collection_type_descriptions = [
                collection_type_descriptions.for_collection_type(t) for t in collection_types
            ]
            # Place higher dimension descriptions first so subcollection mapping
            # (until we expose it to the user) will default to providing tool as much
            # data as possible. So a list:list:paired mapped to a tool that takes
            # list,paired,list:paired - will map over list:paired and create a flat list.
            collection_type_descriptions = sorted(collection_type_descriptions, key=lambda t: t.dimension, reverse=True)
        else:
            collection_type_descriptions = None
        kwargs = dict(collection_type_descriptions=collection_type_descriptions)
        return HistoryQuery(**kwargs)

    @staticmethod
    def from_parameter(param, collection_type_descriptions):
        """Take in a tool parameter element."""
        collection_types = param.collection_types
        return HistoryQuery.from_collection_types(collection_types, collection_type_descriptions)

    def direct_match(self, hdca):
        collection_type_descriptions = self.collection_type_descriptions
        if collection_type_descriptions is not None:
            for collection_type_description in collection_type_descriptions:
                if collection_type_description.can_match_type(hdca.collection.collection_type):
                    return True
            return False

        return True

    def can_map_over(self, hdca):
        collection_type_descriptions = self.collection_type_descriptions
        if collection_type_descriptions is None:
            return False

        hdca_collection_type = hdca.collection.collection_type
        for collection_type_description in collection_type_descriptions:
            # See note about the way this is sorted above.
            if collection_type_description.is_subcollection_of_type(hdca_collection_type):
                return collection_type_description
        return False



# Dan Blankenberg
# Filters a MAF file according to the provided code file, which is generated in maf_filter.xml <configfiles>
# Also allows filtering by number of columns in a block, and limiting output species
from __future__ import print_function

import os
import shutil
import sys

import bx.align.maf

from galaxy.tools.util import maf_utilities


def main():
    # Read command line arguments
    try:
        script_file = sys.argv.pop(1)
        maf_file = sys.argv.pop(1)
        out_file = sys.argv.pop(1)
        additional_files_path = sys.argv.pop(1)
        species = maf_utilities.parse_species_option(sys.argv.pop(1))
        min_size = int(sys.argv.pop(1))
        max_size = int(sys.argv.pop(1))
        if max_size < 1:
            max_size = sys.maxsize
        min_species_per_block = int(sys.argv.pop(1))
        exclude_incomplete_blocks = int(sys.argv.pop(1))
        if species:
            num_species = len(species)
        else:
            num_species = len(sys.argv.pop(1).split(","))
    except Exception:
        print(
            "One or more arguments is missing.\nUsage: maf_filter.py maf_filter_file input_maf output_maf path_to_save_debug species_to_keep",
            file=sys.stderr,
        )
        sys.exit()

    # Open input and output MAF files
    try:
        maf_reader = bx.align.maf.Reader(open(maf_file))
        maf_writer = bx.align.maf.Writer(open(out_file, "w"))
    except Exception:
        print("Your MAF file appears to be malformed.", file=sys.stderr)
        sys.exit()

    # Save script file for debuging/verification info later
    os.mkdir(additional_files_path)
    shutil.copy(script_file, os.path.join(additional_files_path, "debug.txt"))

    # Loop through blocks, running filter on each
    # 'maf_block' and 'ret_val' are used/shared in the provided code file
    # 'ret_val' should be set to True if the block is to be kept
    i = 0
    blocks_kept = 0
    for i, maf_block in enumerate(maf_reader):  # noqa: B007
        if min_size <= maf_block.text_size <= max_size:
            local = {"maf_block": maf_block, "ret_val": False}
            exec(compile(open(script_file).read(), script_file, "exec"), {}, local)
            if local["ret_val"]:
                # Species limiting must be done after filters as filters could be run on non-requested output species
                if species:
                    maf_block = maf_block.limit_to_species(species)
                if len(maf_block.components) >= min_species_per_block and (
                    not exclude_incomplete_blocks or len(maf_block.components) >= num_species
                ):
                    maf_writer.write(maf_block)
                    blocks_kept += 1
    maf_writer.close()
    maf_reader.close()
    if i == 0:
        print("Your file contains no valid maf_blocks.")
    else:
        print("Kept %s of %s blocks (%.2f%%)." % (blocks_kept, i + 1, float(blocks_kept) / float(i + 1) * 100.0))


if __name__ == "__main__":
    main()



"""
Determine if installed tool shed repositories have updates available in their respective tool sheds.
"""

import logging
import threading
from typing import NamedTuple

from sqlalchemy import false

from galaxy import util
from galaxy.model.base import transaction
from galaxy.model.tool_shed_install import ToolShedRepository
from galaxy.tool_shed.util.repository_util import get_tool_shed_status_for_installed_repository
from galaxy.tool_shed.util.shed_util_common import clean_dependency_relationships
from galaxy.util.sleeper import Sleeper
from galaxy.util.tool_shed.common_util import get_tool_shed_url_from_tool_shed_registry
from galaxy.util.tool_shed.encoding_util import tool_shed_decode

log = logging.getLogger(__name__)


class UpdateToChangeset(NamedTuple):
    changeset_revision: str
    ctx_rev: str


class UpdateRepositoryManager:
    def __init__(self, app):
        self.app = app
        self.context = self.app.install_model.context
        # Ideally only one Galaxy server process should be able to check for repository updates.
        if self.app.config.enable_tool_shed_check:
            self.running = True
            self.sleeper = Sleeper()
            self.restarter = threading.Thread(target=self.__restarter)
            self.restarter.daemon = True
            self.app.application_stack.register_postfork_function(self.restarter.start)
            self.seconds_to_sleep = int(app.config.hours_between_check * 3600)

    def get_update_to_changeset_revision_and_ctx_rev(self, repository: ToolShedRepository) -> UpdateToChangeset:
        """Return the changeset revision hash to which the repository can be updated."""
        tool_shed_url = get_tool_shed_url_from_tool_shed_registry(self.app, str(repository.tool_shed))
        params = dict(
            name=str(repository.name),
            owner=str(repository.owner),
            changeset_revision=str(repository.installed_changeset_revision),
        )
        pathspec = ["repository", "get_changeset_revision_and_ctx_rev"]
        try:
            encoded_update_dict = util.url_get(
                tool_shed_url,
                auth=self.app.tool_shed_registry.url_auth(tool_shed_url),
                pathspec=pathspec,
                params=params,
            )
            update_dict = tool_shed_decode(encoded_update_dict)
            changeset_revision = update_dict.get("changeset_revision", None)
            ctx_rev = update_dict.get("ctx_rev", None)
        except Exception as e:
            log.debug(
                f"Error getting change set revision for update from the tool shed for repository '{repository.name}': {str(e)}"
            )
            changeset_revision = None
            ctx_rev = None
        # restore_repository uses changeset_revision, ctx_rev
        return UpdateToChangeset(changeset_revision, ctx_rev)

    def __restarter(self) -> None:
        log.info("Update repository manager restarter starting up...")
        while self.running:
            # Make a call to the Tool Shed for each installed repository to get the latest
            # status information in the Tool Shed for the repository.  This information includes
            # items like newer installable repository revisions, current revision updates, whether
            # the repository revision is the latest installable revision, and whether the repository
            # has been deprecated in the Tool Shed.
            for repository in self.context.query(self.app.install_model.ToolShedRepository).filter(
                self.app.install_model.ToolShedRepository.deleted == false()
            ):
                tool_shed_status_dict = get_tool_shed_status_for_installed_repository(self.app, repository)
                if tool_shed_status_dict:
                    if tool_shed_status_dict != repository.tool_shed_status:
                        repository.tool_shed_status = tool_shed_status_dict
                        with transaction(self.context):
                            self.context.commit()
                else:
                    # The received tool_shed_status_dict is an empty dictionary, so coerce to None.
                    tool_shed_status_dict = None
                    if tool_shed_status_dict != repository.tool_shed_status:
                        repository.tool_shed_status = tool_shed_status_dict
                        with transaction(self.context):
                            self.context.commit()
            self.sleeper.sleep(self.seconds_to_sleep)
        log.info("Update repository manager restarter shutting down...")

    def shutdown(self) -> None:
        if self.app.config.enable_tool_shed_check:
            self.running = False
            self.sleeper.wake()

    def update_repository_record(
        self, repository: ToolShedRepository, updated_metadata_dict, updated_changeset_revision, updated_ctx_rev
    ) -> ToolShedRepository:
        """
        Update a tool_shed_repository database record with new information retrieved from the
        Tool Shed.  This happens when updating an installed repository to a new changeset revision.
        """
        repository.metadata_ = updated_metadata_dict
        tool_shed_url = get_tool_shed_url_from_tool_shed_registry(self.app, repository.tool_shed)
        clean_dependency_relationships(self.app, updated_metadata_dict, repository, tool_shed_url)
        # Update the repository.changeset_revision column in the database.
        repository.changeset_revision = updated_changeset_revision
        repository.ctx_rev = updated_ctx_rev
        # Update the repository.tool_shed_status column in the database.
        tool_shed_status_dict = get_tool_shed_status_for_installed_repository(self.app, repository)
        if tool_shed_status_dict:
            repository.tool_shed_status = tool_shed_status_dict
        else:
            repository.tool_shed_status = None  # type:ignore[assignment]
        session = self.app.install_model.context
        session.add(repository)
        with transaction(session):
            session.commit()
        session.refresh(repository)
        return repository
